from __future__ import annotations
from typing import Any, Dict, Optional, Text, Type
import dataclasses
import uuid

from rasa.engine.caching import Cacheable, TrainingCache
from rasa.engine.graph import ExecutionContext, GraphComponent, SchemaNode
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.engine.training import fingerprinting


class PrecomputedValueProvider(GraphComponent):
    """Holds the precomputed values of a `GraphNode` from a previous training.

    Pre-computed values can either be
    - values loaded from cache
    - values which were provided during the fingerprint run by input nodes
    """

    def __init__(self, output: Cacheable):
        """Initializes a `PrecomputedValueProvider`.

        Args:
            output: The precomputed output to return.
        """
        self._output = output

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> PrecomputedValueProvider:
        """Creates instance (see parent class for full docstring)."""
        return cls(output=config["output"])

    def get_value(self) -> Cacheable:
        """Returns the precomputed output."""
        return self._output

    @classmethod
    def replace_schema_node(cls, node: SchemaNode, output: Any) -> None:
        """Updates a `SchemaNode` to use a `PrecomputedValueProvider`.

        This is for when we want to use the precomputed output value of a node from a
        previous training in a subsequent training. We replace the class in the `uses`
        of the node to a be a `PrecomputedValueProvider` configured to return the
        precomputed value.

        Args:
            node: The node to update.
            output: precomputed cached output that the `PrecomputedValueProvider` will
            return.
        """
        node.uses = cls
        node.config = {"output": output}
        node.fn = cls.get_value.__name__
        node.constructor_name = cls.create.__name__


@dataclasses.dataclass
class FingerprintStatus:
    """Holds the output of a `FingerprintComponent` and is used to prune the graph.

    Attributes:
        output_fingerprint: A fingerprint of the node's output value.
        is_hit: `True` if node's fingerprint key exists in the cache, `False` otherwise.
    """

    output_fingerprint: Optional[Text]
    is_hit: bool

    def fingerprint(self) -> Text:
        """Returns the internal fingerprint.

        If there is no fingerprint returns a random string that will never match.
        """
        return self.output_fingerprint or uuid.uuid4().hex


class FingerprintComponent(GraphComponent):
    """Replaces non-input nodes during a fingerprint run."""

    def __init__(
        self,
        cache: TrainingCache,
        config_of_replaced_component: Dict[Text, Any],
        class_of_replaced_component: Type,
    ) -> None:
        """Initializes a `FingerprintComponent`.

        Args:
            cache: Training cache used to determine if the run is a hit or not.
            config_of_replaced_component: Needed to generate the fingerprint key.
            class_of_replaced_component: Needed to generate the fingerprint key.
        """
        self._cache = cache
        self._config_of_replaced_component = config_of_replaced_component
        self._class_of_replaced_component = class_of_replaced_component

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> FingerprintComponent:
        """Creates a `FingerprintComponent` (see parent class for full docstring)."""
        return cls(
            cache=config["cache"],
            config_of_replaced_component=config["config_of_replaced_component"],
            class_of_replaced_component=config["graph_component_class"],
        )

    def run(self, **kwargs: Any) -> FingerprintStatus:
        """Calculates the fingerprint key to determine if cached output can be used.

        If the fingerprint key matches an entry in the cache it means that there has
        been a previous node execution which matches the same component class, component
        config and input values. This means that we can potentially prune this node
        from the schema, or replace it with a cached value before the next graph run.

        Args:
            **kwargs: Inputs from all parent nodes.

        Returns:
            A `FingerprintStatus` determining if the run was a hit, and if it was a hit
            also the output fingerprint from the cache.
        """
        fingerprint_key = fingerprinting.calculate_fingerprint_key(
            graph_component_class=self._class_of_replaced_component,
            config={
                **self._class_of_replaced_component.get_default_config(),
                **self._config_of_replaced_component,
            },
            inputs=kwargs,
        )

        output_fingerprint = self._cache.get_cached_output_fingerprint(fingerprint_key)

        return FingerprintStatus(
            is_hit=output_fingerprint is not None, output_fingerprint=output_fingerprint
        )

    @classmethod
    def replace_schema_node(cls, node: SchemaNode, cache: TrainingCache) -> None:
        """Updates a `SchemaNode` to use a `FingerprintComponent`.

        This is for when we want to do a fingerprint run. During the fingerprint run we
        replace all non-input nodes with `FingerprintComponent`s so we can determine
        whether they are able to be pruned or cached before the next graph run without
        running the actual components.


        Args:
            node: The node to update.
            cache: The cache is needed to determine of there is cache hit for the
                fingerprint key.
        """
        graph_component_class = node.uses
        node.uses = cls
        # We update the node to be "eager" so that `FingerprintComponent.run` sees
        # ALL the inputs to the node. If it was not eager, we would miss any args used
        # by the constructor.
        node.eager = True
        node.constructor_name = cls.create.__name__
        node.fn = cls.run.__name__
        node.config = {
            "config_of_replaced_component": node.config,
            "cache": cache,
            "graph_component_class": graph_component_class,
        }



import argparse
from typing import List, Optional, Tuple
from urllib.parse import ParseResult, urlparse

import questionary
from rasa.cli import SubParsersAction

import rasa.shared.utils.cli
import rasa.cli.studio.download
import rasa.cli.studio.train
import rasa.cli.studio.upload
from rasa.studio.auth import StudioAuth
from rasa.studio.config import StudioConfig


DEFAULT_REALM_NAME = "rasa-studio"

DEFAULT_CLIENT_ID = "admin-cli"


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    studio_parser = subparsers.add_parser(
        "studio",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Rasa Studio commands.",
    )
    studio_subparsers = studio_parser.add_subparsers()

    rasa.cli.studio.train.add_subparser(studio_subparsers, parents)
    rasa.cli.studio.upload.add_subparser(studio_subparsers, parents)
    rasa.cli.studio.download.add_subparser(studio_subparsers, parents)

    _add_config_subparser(studio_subparsers, parents)
    _add_login_subparser(studio_subparsers, parents)


def _add_config_subparser(
    studio_sub_parsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    studio_config_parser = studio_sub_parsers.add_parser(
        "config",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Configure communication parameters for Rasa Studio",
    )

    studio_config_parser.set_defaults(func=create_and_store_studio_config)

    studio_config_parser.add_argument(
        "--disable-verify",
        "-x",
        action="store_true",
        default=False,
        help="Disable strict SSL verification for the "
        "Rasa Studio authentication server.",
    )

    # add advanced configuration flag to trigger
    # advanced configuration setup for authentication settings
    studio_config_parser.add_argument(
        "--advanced",
        action="store_true",
        default=False,
        help="Configure additional authentication parameters for Rasa Studio",
    )


def _add_login_subparser(
    studio_sub_parsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    login_parser = studio_sub_parsers.add_parser(
        "login",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Login to Rasa Studio",
    )
    login_parser.set_defaults(func=_studio_login)

    login_parser.add_argument(
        "--username",
        type=str,
        help="Username for Rasa Studio",
    )

    login_parser.add_argument(
        "--password",
        type=str,
        help="Password for Rasa Studio",
    )


def _studio_login(args: argparse.Namespace) -> None:
    """Login to Rasa Studio.

    Args:
        args: Commandline arguments
    """
    studio_config = StudioConfig.read_config()
    studio_auth = StudioAuth(studio_config)

    # show the user the studio url they are logging into
    # urlparse will always return parseresult if a string is passed in as url
    parsed_url: ParseResult = urlparse(studio_config.studio_url)  # type: ignore[assignment]
    studio_url = f"{parsed_url.scheme}://{parsed_url.netloc}"

    rasa.shared.utils.cli.print_info(
        f"Trying to log in to Rasa Studio at {studio_url} ..."
    )

    try:
        username = args.username
        password = args.password

        if not username:
            username = questionary.text("Please enter your username").unsafe_ask()

        if not password:
            password = questionary.password("Please enter your password").unsafe_ask()

        studio_auth.login(username, password)
    except KeyboardInterrupt:
        rasa.shared.utils.cli.print_error("Login to Rasa Studio aborted.")
    except Exception as e:
        rasa.shared.utils.cli.print_error(f"Failed to login to Rasa Studio. {e}")


def _configure_studio_url() -> Optional[str]:
    """Configure the Rasa Studio URL.

    Returns:
        The configured Rasa Studio URL
    """
    studio_url = questionary.text(
        "Please provide the Rasa Studio URL",
    ).unsafe_ask()

    if not studio_url.endswith("/"):
        studio_url += "/"

    # prepend with https if no protocol is provided
    if not studio_url.startswith("http"):
        return "https://" + studio_url
    return studio_url


def _get_advanced_config(studio_url: str) -> Tuple:
    """Get the advanced configuration values for Rasa Studio."""
    keycloak_url = questionary.text(
        "Please provide your Rasa Studio Keycloak URL",
        default=studio_url + "auth/",
    ).unsafe_ask()

    realm_name = questionary.text(
        "Please enter Realm Name", default=DEFAULT_REALM_NAME
    ).unsafe_ask()

    client_id = questionary.text(
        "Please enter client ID", default=DEFAULT_CLIENT_ID
    ).unsafe_ask()

    return keycloak_url, realm_name, client_id


def _get_default_config(studio_url: str) -> Tuple:
    """Get the default configuration values for Rasa Studio."""
    keycloak_url = studio_url + "auth/"
    realm_name = DEFAULT_REALM_NAME
    client_id = DEFAULT_CLIENT_ID

    rasa.shared.utils.cli.print_info(
        f"Using default values for "
        f"Keycloak URL: {keycloak_url}, "
        f"Realm Name: '{realm_name}', "
        f"Client ID: '{client_id}'. "
        f"SSL verification is enabled."
        f"You can use '--advanced' to configure these settings."
    )

    return keycloak_url, realm_name, client_id


def _create_studio_config(
    studio_url: str,
    keycloak_url: str,
    realm_name: str,
    client_id: str,
    disable_verify: bool = False,
) -> StudioConfig:
    """Create a StudioConfig object with the provided parameters."""
    return StudioConfig(
        authentication_server_url=keycloak_url,
        studio_url=studio_url + "api/graphql/",
        client_id=client_id,
        realm_name=realm_name,
        disable_verify=disable_verify,
    )


def _check_studio_auth(studio_auth: StudioAuth) -> bool:
    """Check if the Rasa Studio authentication server is reachable."""
    if studio_auth.health_check():
        rasa.shared.utils.cli.print_info(
            "Tried configuration and successfully reached Rasa Studio."
        )
        return True
    return False


def _prompt_store_config_anyways() -> bool:
    """Prompt the user to store the configuration."""
    should_store_config = questionary.confirm(
        "Do you want to store the configuration anyway?"
    ).unsafe_ask()

    return should_store_config


def _configure_studio_config(args: argparse.Namespace) -> StudioConfig:
    """Configure the Rasa Studio connection settings."""
    studio_url = _configure_studio_url()

    # check if the user wants to configure advanced authentication settings
    if args.advanced:
        keycloak_url, realm_name, client_id = _get_advanced_config(studio_url)
    else:
        keycloak_url, realm_name, client_id = _get_default_config(studio_url)

    # create a configuration and auth object to try to reach the studio
    studio_config = _create_studio_config(
        studio_url,
        keycloak_url,
        realm_name,
        client_id,
        disable_verify=args.disable_verify,
    )

    if studio_config.disable_verify:
        rasa.shared.utils.cli.print_info(
            "Disabling SSL verification for the Rasa Studio authentication server."
        )
    else:
        rasa.shared.utils.cli.print_info(
            "Enabling SSL verification for the Rasa Studio authentication server."
        )

    studio_auth = StudioAuth(studio_config)

    if _check_studio_auth(studio_auth):
        return studio_config

    rasa.shared.utils.cli.print_error(
        f"Failed to reach Rasa Studio authentication server at "
        f"{studio_config.authentication_server_url}."
    )

    if _prompt_store_config_anyways():
        return studio_config

    rasa.shared.utils.cli.print_info(
        "The config was not stored, you can change the values "
        "again or abort with Ctrl-C."
    )

    return _configure_studio_config(args)


def create_and_store_studio_config(args: argparse.Namespace) -> None:
    """Create and store the Rasa Studio configuration.

    Args:
        args: Commandline arguments
    """
    try:
        rasa.shared.utils.cli.print_info(
            "Configuring Rasa Studio connection settings ..."
        )
        studio_config = _configure_studio_config(args)

        if studio_config:
            studio_config.write_config()

            rasa.shared.utils.cli.print_success(
                "Successfully configured Rasa Pro to connect to Studio."
            )
        else:
            rasa.shared.utils.cli.print_error(
                "Failed to configure Rasa Pro to connect to Studio."
            )
    except KeyboardInterrupt:
        rasa.shared.utils.cli.print_error("Configuration of Rasa Studio aborted.")



from typing import List, Optional

from rasa.core.channels import UserMessage
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.core.utils import AvailableEndpoints

PLACEHOLDER_IMPORTER = "__importer__"
PLACEHOLDER_MESSAGE = "__message__"
PLACEHOLDER_TRACKER = "__tracker__"
PLACEHOLDER_ENDPOINTS = "__endpoints__"
RESERVED_PLACEHOLDERS = {
    PLACEHOLDER_IMPORTER: TrainingDataImporter,
    PLACEHOLDER_MESSAGE: List[UserMessage],
    PLACEHOLDER_TRACKER: DialogueStateTracker,
    PLACEHOLDER_ENDPOINTS: Optional[AvailableEndpoints],
}



from __future__ import annotations
from abc import abstractmethod, ABC
from collections import Counter
from typing import Generic, Iterable, Text, Optional, Dict, Any, TypeVar

from rasa.nlu.constants import FEATURIZER_CLASS_ALIAS
from rasa.shared.nlu.training_data.features import Features
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.exceptions import InvalidConfigException
from rasa.shared.nlu.constants import FEATURE_TYPE_SENTENCE, FEATURE_TYPE_SEQUENCE

FeatureType = TypeVar("FeatureType")


class Featurizer(Generic[FeatureType], ABC):
    """Base class for all featurizers."""

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the component's default config."""
        return {FEATURIZER_CLASS_ALIAS: None}

    def __init__(self, name: Text, config: Dict[Text, Any]) -> None:
        """Instantiates a new featurizer.

        Args:
          config: configuration
          name: a name that can be used as identifier, in case the configuration does
            not specify an `alias` (or this `alias` is None)
        """
        super().__init__()
        self.validate_config(config)
        self._config = config
        self._identifier = self._config[FEATURIZER_CLASS_ALIAS] or name

    @classmethod
    @abstractmethod
    def validate_config(cls, config: Dict[Text, Any]) -> None:
        """Validates that the component is configured properly."""
        ...

    def add_features_to_message(
        self,
        sequence: FeatureType,
        sentence: Optional[FeatureType],
        attribute: Text,
        message: Message,
    ) -> None:
        """Adds sequence and sentence features for the attribute to the given message.

        Args:
          sequence: sequence feature matrix
          sentence: sentence feature matrix
          attribute: the attribute which both features describe
          message: the message to which we want to add those features
        """
        for type, features in [
            (FEATURE_TYPE_SEQUENCE, sequence),
            (FEATURE_TYPE_SENTENCE, sentence),
        ]:
            if features is not None:
                wrapped_feature = Features(features, type, attribute, self._identifier)
                message.add_features(wrapped_feature)

    @staticmethod
    def raise_if_featurizer_configs_are_not_compatible(
        featurizer_configs: Iterable[Dict[Text, Any]],
    ) -> None:
        """Validates that the given configurations of featurizers can be used together.

        Raises:
          `InvalidConfigException` if the given featurizers should not be used in
            the same graph.
        """
        # NOTE: this assumes the names given via the execution context are unique
        alias_counter = Counter(
            config[FEATURIZER_CLASS_ALIAS]
            for config in featurizer_configs
            if FEATURIZER_CLASS_ALIAS in config
        )
        if not alias_counter:  # no alias found
            return
        if alias_counter.most_common(1)[0][1] > 1:
            raise InvalidConfigException(
                f"Expected the featurizers to have unique names but found "
                f" (name, count): {alias_counter.most_common()}. "
                f"Please update your config such that each featurizer has a unique "
                f"alias."
            )



from __future__ import annotations
import dataclasses
import logging
from typing import Dict, Text, Any

import rasa.shared.utils.io
from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class RuleOnlyDataProvider(GraphComponent):
    """Provides slots and loops that are only used in rules to other policies.

    Policies can use this to exclude features which are only used by rules from the
    featurization.
    """

    rule_only_data: Dict[Text, Any]

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> RuleOnlyDataProvider:
        """Creates component (see parent class for docstring)."""
        rule_only_data = {}
        try:
            with model_storage.read_from(resource) as directory:
                rule_only_data = rasa.shared.utils.io.read_json_file(
                    directory / "rule_only_data.json"
                )
        except ValueError:
            logger.debug(
                "Failed to load rule-only data from a trained 'RulePolicy'. "
                "Providing empty rule-only data instead."
            )

        return cls(rule_only_data)

    def provide(self) -> Dict[Text, Any]:
        """Provides data to other graph component."""
        return self.rule_only_data



import json
from typing import Text, Optional, Dict, Any

import aiohttp
import logging
from sanic.exceptions import SanicException
import jwt
import jwt.exceptions

import rasa.core.channels.channel
from rasa.core.channels.channel import InputChannel
from rasa.core.channels.rest import RestInput
from rasa.core.constants import DEFAULT_REQUEST_TIMEOUT
from sanic.request import Request

logger = logging.getLogger(__name__)

CONVERSATION_ID_KEY = "conversation_id"
JWT_USERNAME_KEY = "username"
INTERACTIVE_LEARNING_PERMISSION = "clientEvents:create"


class RasaChatInput(RestInput):
    """Chat input channel for Rasa Enterprise."""

    @classmethod
    def name(cls) -> Text:
        """Name of the channel."""
        return "rasa"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        if not credentials:
            cls.raise_missing_credentials_exception()

        return cls(credentials.get("url"))

    def __init__(self, url: Optional[Text]) -> None:
        """Initialise the channel with attributes."""
        self.base_url = url
        self.jwt_key: Optional[Text] = None
        self.jwt_algorithm = None

    async def _fetch_public_key(self) -> None:
        public_key_url = f"{self.base_url}/version"
        async with aiohttp.ClientSession() as session:
            async with session.get(
                public_key_url, timeout=DEFAULT_REQUEST_TIMEOUT
            ) as resp:
                status_code = resp.status
                if status_code != 200:
                    logger.error(
                        "Failed to fetch JWT public key from URL '{}' with "
                        "status code {}: {}"
                        "".format(public_key_url, status_code, await resp.text())
                    )
                    return
                rjs = await resp.json()
                public_key_field = "keys"
                if public_key_field in rjs:
                    self.jwt_key = rjs["keys"][0]["key"]
                    self.jwt_algorithm = rjs["keys"][0]["alg"]
                    logger.debug(
                        "Fetched JWT public key from URL '{}' for algorithm '{}':\n{}"
                        "".format(public_key_url, self.jwt_algorithm, self.jwt_key)
                    )
                else:
                    logger.error(
                        "Retrieved json response from URL '{}' but could not find "
                        "'{}' field containing the JWT public key. Please make sure "
                        "you use an up-to-date version of Rasa Enterprise (>= 0.20.2). "
                        "Response was: {}"
                        "".format(public_key_url, public_key_field, json.dumps(rjs))
                    )

    async def _decode_bearer_token(self, bearer_token: Text) -> Optional[Dict]:
        if self.jwt_key is None:
            await self._fetch_public_key()

        try:
            return rasa.core.channels.channel.decode_jwt(
                bearer_token, self.jwt_key, self.jwt_algorithm
            )
        except jwt.InvalidSignatureError:
            logger.error("JWT public key invalid, fetching new one.")
            await self._fetch_public_key()
            return rasa.core.channels.channel.decode_jwt(
                bearer_token, self.jwt_key, self.jwt_algorithm
            )

    async def _extract_sender(self, req: Request) -> Optional[Text]:
        """Fetch user from the Rasa Enterprise Admin API."""
        jwt_payload = None
        if req.headers.get("Authorization"):
            jwt_payload = await self._decode_bearer_token(req.headers["Authorization"])

        if not jwt_payload:
            jwt_payload = await self._decode_bearer_token(req.args.get("token"))

        if not jwt_payload:
            raise SanicException(status_code=401)

        if CONVERSATION_ID_KEY in req.json:
            if self._has_user_permission_to_send_messages_to_conversation(
                jwt_payload, req.json
            ):
                return req.json[CONVERSATION_ID_KEY]
            else:
                logger.error(
                    "User '{}' does not have permissions to send messages to "
                    "conversation '{}'.".format(
                        jwt_payload[JWT_USERNAME_KEY], req.json[CONVERSATION_ID_KEY]
                    )
                )
                raise SanicException(status_code=401)

        return jwt_payload[JWT_USERNAME_KEY]

    @staticmethod
    def _has_user_permission_to_send_messages_to_conversation(
        jwt_payload: Dict, message: Dict
    ) -> bool:
        user_scopes = jwt_payload.get("scopes", [])
        return INTERACTIVE_LEARNING_PERMISSION in user_scopes or message[
            CONVERSATION_ID_KEY
        ] == jwt_payload.get(JWT_USERNAME_KEY)



from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Optional,
    Text,
)

import structlog
from rasa.core.channels import SocketIOInput
from sanic import response

if TYPE_CHECKING:
    from rasa.core.channels.channel import InputChannel, UserMessage
    from sanic import Blueprint
    from sanic.request import Request
    from sanic.response import HTTPResponse


INSPECT_TEMPLATE_PATH = "inspector/dist"

structlogger = structlog.get_logger()


class DevelopmentInspectInput(SocketIOInput):
    """Internal Rasa Pro-only SocketIO-based channel.

    It extends the Rasa Pro SocketIO channel by
    adding the /inspect endpoint, which works hand in hand with `rasa inspect`.
    """

    @classmethod
    def name(cls) -> Text:
        """Channel name."""
        return "inspector"

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> "InputChannel":
        """Override function to set default parameters."""
        credentials = credentials or {}
        if credentials.get("session_persistence") is False:
            structlogger.error("development_inspector.init.session_persistence_ignored")

        # session_persistence needs to be True
        credentials["session_persistence"] = True
        return super().from_credentials(credentials)

    def __init__(
        self,
        user_message_evt: Text = "user_uttered",
        bot_message_evt: Text = "bot_uttered",
        namespace: Optional[Text] = None,
        # this argument is kept to keep the signature compatible with the parent class
        # especially with from_credentials()
        session_persistence: bool = True,
        socketio_path: Optional[Text] = "/socket.io",
        jwt_key: Optional[Text] = None,
        jwt_method: Optional[Text] = "HS256",
        metadata_key: Optional[Text] = "metadata",
    ):
        """Override function to set default parameters."""
        super().__init__(
            user_message_evt,
            bot_message_evt,
            namespace,
            True,  # session_persistence needs to be True
            socketio_path,
            jwt_key,
            jwt_method,
            metadata_key,
        )

    @staticmethod
    def inspect_html_path() -> Text:
        """Returns the path to the inspect.html file."""
        import pkg_resources

        return pkg_resources.resource_filename(__name__, INSPECT_TEMPLATE_PATH)

    def blueprint(
        self, on_new_message: Callable[["UserMessage"], Awaitable[Any]]
    ) -> "Blueprint":
        """Defines a Sanic blueprint."""
        socketio_webhook: "Blueprint" = super().blueprint(on_new_message)
        socketio_webhook.static("/assets", self.inspect_html_path() + "/assets")

        @socketio_webhook.route("/inspect.html", methods=["GET"])
        async def inspect(_: "Request") -> "HTTPResponse":
            return await response.file(self.inspect_html_path() + "/index.html")

        return socketio_webhook



import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional, Text

from rasa.shared.exceptions import RasaException, YamlException
from rasa.shared.utils.yaml import (
    read_config_file,
    read_yaml_file,
    validate_yaml_content_using_schema,
)

logger = logging.getLogger(__name__)
SCHEMA_FILE = Path(__file__).parent / "schemas" / "config.yml"


def read_endpoint_config(
    filename: Optional[Text], endpoint_type: Text
) -> Optional[Dict[Text, Any]]:
    """Reads a configuration file from disk and extract one config."""
    if not filename:
        return None

    try:
        content = read_config_file(filename, reader_type=["safe", "rt"])

        if content.get(endpoint_type) is None:
            return None

        return content
    except FileNotFoundError:
        logger.error(
            "Failed to read configuration from {}. No such file.".format(
                os.path.abspath(filename)
            ),
        )
        return None


def extract_anonymization_traits(
    anonymization_config: Optional[Dict[Text, Any]], endpoint_type: Text
) -> Dict[Text, Any]:
    """Extract anonymization traits from the anonymization config.

    Args:
        anonymization_config: The anonymization config.
        endpoint_type: The endpoint type.

    Returns:
        A dictionary containing the anonymization traits.
    """
    if not anonymization_config:
        return {"enabled": False}

    anonymization_value = anonymization_config.get(endpoint_type)
    if not anonymization_value:
        logger.debug("Anonymization config found but nothing is defined.")
        return {"enabled": False}

    try:
        validate_anonymization_yaml(anonymization_config)
    except RasaException as exception:
        logger.debug(exception)
        return {"enabled": False}

    rule_lists = anonymization_value.get("rule_lists", [])

    traits = {
        "enabled": True,
        "metadata": anonymization_value.get("metadata", {}),
        "number_of_rule_lists": len(rule_lists),
        "number_of_rules": 0,
        "substitutions": {
            "mask": 0,
            "faker": 0,
            "text": 0,
            "not_defined": 0,
        },
        "entities": set(),
    }

    for rule_list in rule_lists:
        traits["number_of_rules"] += len(rule_list.get("rules", []))
        for rule in rule_list.get("rules", []):
            traits["substitutions"][rule.get("substitution", "not_defined")] += 1
            traits["entities"].add(rule.get("entity"))

    entities = list(traits["entities"])
    entities.sort()
    traits["entities"] = entities

    return traits


def validate_anonymization_yaml(yaml_content: Dict[Text, Any]) -> None:
    """Checks if the yaml_content adheres to the anonymization rule schema.

    If the yaml_content is not in the right format, an exception will be raised.
    """
    schema = read_yaml_file(SCHEMA_FILE, reader_type=("safe", "rt"))
    try:
        validate_yaml_content_using_schema(yaml_content, schema)
    except YamlException as exception:
        raise RasaException(
            f"Invalid configuration for `anonymization_rules` : {exception}"
        ) from exception

    rule_lists = yaml_content.get("anonymization", {}).get("rule_lists", [])

    rule_set = set()
    for rule_list in rule_lists:
        rule_id = rule_list.get("id")
        if rule_id in rule_set:
            raise RasaException(
                f"Invalid configuration for `anonymization_rules` :"
                f"Duplicate rule id: '{rule_id}' encountered in rule_lists"
            )
        rule_set.add(rule_id)



import argparse
from pathlib import Path
from typing import List, Text

from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import add_domain_param
from rasa.shared.core.domain import Domain
from rasa.shared.utils.cli import print_error_and_exit

from rasa.markers.upload import upload


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add all marker parsers.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    markers_parser = subparsers.add_parser(
        "markers",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Rasa Studio commands.",
    )
    marker_subparser = markers_parser.add_subparsers()

    _add_upload_subparser(marker_subparser, parents)


def _add_upload_subparser(
    marker_sub_parsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    marker_upload_sub_parser = marker_sub_parsers.add_parser(
        "upload",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Upload Markers to Rasa Pro Services",
    )

    marker_upload_sub_parser.set_defaults(func=_upload_markers)

    marker_upload_sub_parser.add_argument(
        "--config",
        default="markers.yml",
        type=Path,
        help="The marker configuration file(s) containing marker definitions. "
        "This can be a single YAML file, or a directory that contains several "
        "files with marker definitions in it. The content of these files will "
        "be read and merged together.",
    )

    marker_upload_sub_parser.add_argument(
        "--rasa-pro-services-url",
        default="",
        type=Text,
        help="The URL of the Rasa Pro Services instance to upload markers to."
        "Specified URL should not contain a trailing slash.",
    )

    add_domain_param(marker_upload_sub_parser)


def _upload_markers(args: argparse.Namespace) -> None:
    markers_path = args.config
    domain_path = args.domain
    url = args.rasa_pro_services_url

    domain = Domain.load(domain_path) if domain_path else None

    if domain is None:
        print_error_and_exit("No domain specified. Skipping validation.")

    upload(url=url, domain=domain, markers_path=markers_path)



ENTITIES_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "start": {"type": "integer"},
            "end": {"type": "integer"},
            "entity": {"type": "string"},
            "confidence": {"type": "number"},
            "extractor": {"type": ["string", "null"]},
            "value": {},
            "role": {"type": ["string", "null"]},
            "group": {"type": ["string", "null"]},
        },
        "required": ["entity", "value"],
    },
}

INTENT = {
    "type": "object",
    "properties": {"name": {"type": "string"}, "confidence": {"type": "number"}},
}

RESPONSE_SCHEMA = {
    "type": "object",
    "properties": {
        "responses": {
            "type": "array",
            "items": {"type": "object", "properties": {"text": {"type": "string"}}},
        },
        "response_templates": {
            "type": "array",
            "items": {"type": "object", "properties": {"text": {"type": "string"}}},
        },
        "confidence": {"type": "number"},
        "intent_response_key": {"type": "string"},
        "utter_action": {"type": "string"},
        "template_name": {"type": "string"},
    },
}

RANKING_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "id": {"type": "number"},
            "confidence": {"type": "number"},
            "intent_response_key": {"type": "string"},
        },
    },
}

USER_UTTERED = {
    "properties": {
        "event": {"const": "user"},
        "text": {"type": ["string", "null"]},
        "input_channel": {"type": ["string", "null"]},
        "message_id": {"type": ["string", "null"]},
        "parse_data": {
            "type": "object",
            "properties": {
                "text": {"type": ["string", "null"]},
                "intent_ranking": {"type": "array", "items": INTENT},
                "intent": INTENT,
                "entities": ENTITIES_SCHEMA,
                "response_selector": {
                    "type": "object",
                    "oneOf": [
                        {"properties": {"all_retrieval_intents": {"type": "array"}}},
                        {
                            "patternProperties": {
                                "[\\w/]": {
                                    "type": "object",
                                    "properties": {
                                        "response": RESPONSE_SCHEMA,
                                        "ranking": RANKING_SCHEMA,
                                    },
                                }
                            }
                        },
                    ],
                },
            },
        },
    }
}

ACTION_EXECUTED = {
    "properties": {
        "event": {"const": "action"},
        "policy": {"type": ["string", "null"]},
        "confidence": {"type": ["number", "null"]},
        "name": {"type": ["string", "null"]},
        "hide_rule_turn": {"type": "boolean"},
        "action_text": {"type": ["string", "null"]},
    }
}

SLOT_SET = {
    "properties": {"event": {"const": "slot"}, "name": {"type": "string"}, "value": {}},
    "required": ["name", "value"],
}

ENTITIES_ADDED = {
    "properties": {"event": {"const": "entities"}, "entities": ENTITIES_SCHEMA},
    "required": ["entities"],
}

USER_UTTERED_FEATURIZATION = {"properties": {"event": {"const": "user_featurization"}}}
REMINDER_CANCELLED = {"properties": {"event": {"const": "cancel_reminder"}}}
REMINDER_SCHEDULED = {"properties": {"event": {"const": "reminder"}}}
ACTION_EXECUTION_REJECTED = {
    "properties": {"event": {"const": "action_execution_rejected"}}
}
FORM_VALIDATION = {"properties": {"event": {"const": "form_validation"}}}
LOOP_INTERRUPTED = {"properties": {"event": {"const": "loop_interrupted"}}}
FORM = {"properties": {"event": {"const": "form"}}}
ACTIVE_LOOP = {"properties": {"event": {"const": "active_loop"}}}
ALL_SLOTS_RESET = {"properties": {"event": {"const": "reset_slots"}}}
CONVERSATION_RESUMED = {"properties": {"event": {"const": "resume"}}}
CONVERSATION_PAUSED = {"properties": {"event": {"const": "pause"}}}
FOLLOWUP_ACTION = {"properties": {"event": {"const": "followup"}}}
STORY_EXPORTED = {"properties": {"event": {"const": "export"}}}
RESTARTED = {"properties": {"event": {"const": "restart"}}}
ACTION_REVERTED = {"properties": {"event": {"const": "undo"}}}
USER_UTTERANCE_REVERTED = {"properties": {"event": {"const": "rewind"}}}
BOT_UTTERED = {"properties": {"event": {"const": "bot"}}}
SESSION_STARTED = {"properties": {"event": {"const": "session_started"}}}
AGENT_UTTERED = {"properties": {"event": {"const": "agent"}}}
FLOW_STARTED = {
    "properties": {"event": {"const": "flow_started"}, "flow_id": {"type": "string"}}
}
FLOW_INTERRUPTED = {
    "properties": {
        "event": {"const": "flow_interrupted"},
        "flow_id": {"type": "string"},
        "step_id": {"type": "string"},
    }
}
FLOW_RESUMED = {
    "properties": {
        "event": {"const": "flow_resumed"},
        "flow_id": {"type": "string"},
        "step_id": {"type": "string"},
    }
}
FLOW_COMPLETED = {
    "properties": {
        "event": {"const": "flow_completed"},
        "flow_id": {"type": "string"},
        "step_id": {"type": "string"},
    }
}
FLOW_CANCELLED = {
    "properties": {
        "event": {"const": "flow_cancelled"},
        "flow_id": {"type": "string"},
        "step_id": {"type": "string"},
    }
}
DIALOGUE_STACK_UPDATED = {
    "properties": {"event": {"const": "stack"}, "update": {"type": "array"}}
}
ROUTING_SESSION_ENDED = {"properties": {"event": {"const": "routing_session_ended"}}}

ERROR_HANDLED = {"properties": {"event": {"const": "error"}}}

EVENT_SCHEMA = {
    "type": "object",
    "properties": {
        "event": {"type": "string"},
        "timestamp": {"type": ["number", "null"]},
        "metadata": {"type": ["object", "null"]},
    },
    "required": ["event"],
    "oneOf": [
        USER_UTTERED,
        ACTION_EXECUTED,
        SLOT_SET,
        ENTITIES_ADDED,
        USER_UTTERED_FEATURIZATION,
        REMINDER_CANCELLED,
        REMINDER_SCHEDULED,
        ACTION_EXECUTION_REJECTED,
        FORM_VALIDATION,
        LOOP_INTERRUPTED,
        FORM,
        ACTIVE_LOOP,
        ALL_SLOTS_RESET,
        CONVERSATION_RESUMED,
        CONVERSATION_PAUSED,
        FOLLOWUP_ACTION,
        STORY_EXPORTED,
        RESTARTED,
        ACTION_REVERTED,
        USER_UTTERANCE_REVERTED,
        BOT_UTTERED,
        SESSION_STARTED,
        AGENT_UTTERED,
        FLOW_STARTED,
        FLOW_INTERRUPTED,
        FLOW_RESUMED,
        FLOW_COMPLETED,
        FLOW_CANCELLED,
        DIALOGUE_STACK_UPDATED,
        ROUTING_SESSION_ENDED,
        ERROR_HANDLED,
    ],
}

EVENTS_SCHEMA = {"type": "array", "items": EVENT_SCHEMA}



from rasa.dialogue_understanding.stack.frames.dialogue_stack_frame import (
    DialogueStackFrame,
)
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    UserFlowStackFrame,
    BaseFlowStackFrame,
)
from rasa.dialogue_understanding.stack.frames.pattern_frame import PatternFlowStackFrame
from rasa.dialogue_understanding.stack.frames.search_frame import SearchStackFrame
from rasa.dialogue_understanding.stack.frames.chit_chat_frame import ChitChatStackFrame

__all__ = [
    "DialogueStackFrame",
    "BaseFlowStackFrame",
    "PatternFlowStackFrame",
    "UserFlowStackFrame",
    "SearchStackFrame",
    "ChitChatStackFrame",
]



import logging

import rasa

from rasa.core.nlg.contextual_response_rephraser import ContextualResponseRephraser
from rasa.core.policies.enterprise_search_policy import EnterpriseSearchPolicy
from rasa.core.policies.intentless_policy import IntentlessPolicy

logging.getLogger(__name__).addHandler(logging.NullHandler())

__version__ = rasa.__version__

__all__ = [
    "EnterpriseSearchPolicy",
    "IntentlessPolicy",
    "ContextualResponseRephraser",
]



import asyncio
import logging

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from pytz import UnknownTimeZoneError, utc
import rasa.shared.utils.io

__scheduler = None

logger = logging.getLogger(__name__)


async def scheduler() -> AsyncIOScheduler:
    """Thread global scheduler to handle all recurring tasks.

    If no scheduler exists yet, this will instantiate one.
    """
    global __scheduler

    if not __scheduler:
        try:
            __scheduler = AsyncIOScheduler(event_loop=asyncio.get_event_loop())
            __scheduler.start()
            return __scheduler
        except UnknownTimeZoneError:
            rasa.shared.utils.io.raise_warning(
                "apscheduler could not find a timezone and is "
                "defaulting to utc. This is probably because "
                "your system timezone is not set. "
                'Set it with e.g. echo "Europe/Berlin" > '
                "/etc/timezone"
            )
            __scheduler = AsyncIOScheduler(
                event_loop=asyncio.get_event_loop(), timezone=utc
            )
            __scheduler.start()
            return __scheduler
    else:
        # scheduler already created, make sure it is running on
        # the correct loop
        # noinspection PyProtectedMember
        if not __scheduler._eventloop == asyncio.get_event_loop():
            raise RuntimeError(
                "Detected inconsistent loop usage. "
                "Trying to schedule a task on a new event "
                "loop, but scheduler was created with a "
                "different event loop. Make sure there "
                "is only one event loop in use and that the "
                "scheduler is running on that one."
            )
        return __scheduler


def kill_scheduler() -> None:
    """Terminate the scheduler if started.

    Another call to `scheduler` will create a new scheduler.
    """
    global __scheduler

    if __scheduler:
        __scheduler.shutdown()
        __scheduler = None



from __future__ import annotations
import copy

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional
import structlog
import typing
import jsonpatch

if typing.TYPE_CHECKING:
    from rasa.dialogue_understanding.stack.frames import DialogueStackFrame

structlogger = structlog.get_logger()


@dataclass
class DialogueStack:
    """Represents the current dialogue stack."""

    frames: List["DialogueStackFrame"]

    @staticmethod
    def from_dict(data: List[Dict[str, Any]]) -> DialogueStack:
        """Creates a `DialogueStack` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStack` from.

        Returns:
            The created `DialogueStack`.
        """
        from rasa.dialogue_understanding.stack.frames import DialogueStackFrame

        return DialogueStack(
            frames=[DialogueStackFrame.create_typed_frame(frame) for frame in data]
        )

    @staticmethod
    def empty() -> DialogueStack:
        """Creates an empty `DialogueStack`.

        Returns:
            The created empty `DialogueStack`.
        """
        return DialogueStack(frames=[])

    def as_dict(self) -> List[Dict[str, Any]]:
        """Returns the `DialogueStack` as a dictionary.

        Returns:
            The `DialogueStack` as a dictionary.
        """
        return [frame.as_dict() for frame in self.frames]

    def copy(self) -> DialogueStack:
        return copy.deepcopy(self)

    def push(self, frame: "DialogueStackFrame", index: Optional[int] = None) -> None:
        """Pushes a new frame onto the stack.

        If the frame shouldn't be put on top of the stack, the index can be
        specified. Not specifying an index equals `push(frame, index=len(frames))`.

        Args:
            frame: The frame to push onto the stack.
            index: The index to insert the frame at. If `None`, the frame
                is put on top of the stack.
        """
        if index is None:
            self.frames.append(frame)
        else:
            self.frames.insert(index, frame)

    def update(self, frame: "DialogueStackFrame") -> None:
        """Updates the topmost frame.

        Args:
            frame: The frame to update.
        """
        if not self.is_empty():
            self.pop()

        self.push(frame)

    def pop(self) -> "DialogueStackFrame":
        """Pops the topmost frame from the stack.

        Returns:
            The popped frame.
        """
        return self.frames.pop()

    def current_context(self) -> Dict[str, Any]:
        """Returns the context of the topmost frame.

        Returns:
            The context of the topmost frame.
        """
        if self.is_empty():
            return {}

        return self.frames[-1].context_as_dict(self.frames[:-1])

    def top(
        self,
        ignore: Optional[Callable[["DialogueStackFrame"], bool]] = None,
    ) -> Optional["DialogueStackFrame"]:
        """Returns the topmost frame from the stack.

        Args:
            ignore_frame: The ID of the flow to ignore. Picks the top most
                frame that has a different flow ID.

        Returns:
            The topmost frame.
        """
        for frame in reversed(self.frames):
            if ignore and ignore(frame):
                continue
            return frame
        return None

    def is_empty(self) -> bool:
        """Checks if the stack is empty.

        Returns:
            `True` if the stack is empty, `False` otherwise.
        """
        return len(self.frames) == 0

    def update_from_patch(self, patch_dump: str) -> DialogueStack:
        """Updates the stack from a patch.

        Args:
            patch_dump: The patch to apply to the stack.

        Returns:
            The updated stack."""
        patch = jsonpatch.JsonPatch.from_string(patch_dump)
        dialogue_stack_dump = patch.apply(self.as_dict())
        return DialogueStack.from_dict(dialogue_stack_dump)

    def create_stack_patch(self, updated_stack: DialogueStack) -> Optional[str]:
        """Creates a patch to update the stack to the updated stack state.

        Example:
            > stack = DialogueStack.from_dict([
            >     {
            >         "type": "flow",
            >         "frame_type": "regular",
            >         "flow_id": "foo",
            >         "step_id": "START",
            >         "frame_id": "test",
            >     }
            > ])
            > updated_stack = DialogueStack.from_dict([
            >     {
            >         "type": "flow",
            >         "frame_type": "regular",
            >         "flow_id": "foo",
            >         "step_id": "1",
            >         "frame_id": "test",
            >     }
            > ])
            > stack.create_stack_patch(updated_stack)
            '[{"op": "replace", "path": "/0/step_id", "value": "1"}]'

        Args:
            updated_stack: The updated stack.

        Returns:
            The patch to update the stack to the updated stack state.
        """
        patch = jsonpatch.JsonPatch.from_diff(self.as_dict(), updated_stack.as_dict())

        if patch:
            return patch.to_string()
        return None



import logging
import typing
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text, Union

from faker import Faker
from presidio_analyzer import AnalyzerEngine
from presidio_analyzer.nlp_engine import (
    SpacyNlpEngine,
    StanzaNlpEngine,
    TransformersNlpEngine,
)
from presidio_anonymizer import AnonymizerEngine
from presidio_anonymizer.entities import OperatorConfig
from rasa.shared.exceptions import RasaException

from rasa.utils.singleton import Singleton

if typing.TYPE_CHECKING:
    from presidio_analyzer.nlp_engine.nlp_engine import NlpEngine

DEFAULT_PRESIDIO_LANG_CODE = "en"
DEFAULT_PRESIDIO_MODEL_NAME = "en_core_web_lg"
DEFAULT_PRESIDIO_MODEL_PROVIDER = "spacy"

logger = logging.getLogger(__name__)


@dataclass
class AnonymizationRule:
    """A rule for anonymizing a given text."""

    entity_name: Text
    substitution: Text = "mask"
    value: Optional[Text] = None


@dataclass
class AnonymizationRuleList:
    """A list of anonymization rules."""

    id: Text
    rule_list: List[AnonymizationRule]
    language: Text = DEFAULT_PRESIDIO_LANG_CODE
    model_provider: Text = DEFAULT_PRESIDIO_MODEL_PROVIDER
    models: Union[Text, Dict[Text, Text]] = DEFAULT_PRESIDIO_MODEL_NAME


class AnonymizationAnalyzer(metaclass=Singleton):
    """Anonymization analyzer."""

    presidio_analyzer_engine = None

    def __init__(self, anonymization_rule_list: AnonymizationRuleList):
        """Initialise the anonymization analyzer."""
        if self.presidio_analyzer_engine is None:
            self.presidio_analyzer_engine = self._get_analyzer_engine(
                anonymization_rule_list
            )

    @staticmethod
    def _get_analyzer_engine(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> AnalyzerEngine:
        """Returns an analyzer engine for all the anonymization rule lists."""
        try:
            nlp_engine = AnonymizationAnalyzer._build_presidio_nlp_engine(
                anonymization_rule_list
            )
        except (OSError, ImportError) as exception:
            raise RasaException(
                "Failed to load Presidio nlp engine. "
                "Please check that you have provided "
                "a valid model name."
            ) from exception

        return AnalyzerEngine(
            nlp_engine=nlp_engine,
            supported_languages=[anonymization_rule_list.language],
        )

    @staticmethod
    def _build_presidio_nlp_engine(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> "NlpEngine":
        """Creates an instance of the Presidio nlp engine."""
        if anonymization_rule_list.model_provider == "transformers":
            nlp_engine = TransformersNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )
        elif anonymization_rule_list.model_provider == "stanza":
            nlp_engine = StanzaNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )
        else:
            nlp_engine = SpacyNlpEngine(
                models={
                    anonymization_rule_list.language: anonymization_rule_list.models
                },
            )

        return nlp_engine


class AnonymizationRuleExecutor:
    """Executes a given anonymization rule set on a given text."""

    def __init__(self, anonymization_rule_list: AnonymizationRuleList):
        """Initialize the anonymization rule executor."""
        self.anonymization_rule_list = anonymization_rule_list

        is_valid_rule_list = self._validate_anonymization_rule_list(
            anonymization_rule_list
        )

        self.analyzer = (
            AnonymizationAnalyzer(anonymization_rule_list)
            if is_valid_rule_list
            else None
        )

        self.anonymizer_engine = AnonymizerEngine()  # type: ignore

    @staticmethod
    def _validate_anonymization_rule_list(
        anonymization_rule_list: AnonymizationRuleList,
    ) -> bool:
        """Validates the given anonymization rule list object."""
        if (
            anonymization_rule_list.language != DEFAULT_PRESIDIO_LANG_CODE
            and anonymization_rule_list.models == DEFAULT_PRESIDIO_MODEL_NAME
        ):
            logger.debug(
                f"Anonymization rule list language is "
                f"'{anonymization_rule_list.language}', "
                f"but no specific model name was provided. "
                f"You must specify the spaCy model name in the"
                f"endpoints yaml file. "
                f"Cannot proceed with anonymization."
            )
            return False

        return True

    def run(self, text: Text) -> Optional[Text]:
        """Anonymizes the given text using the given anonymization rule list."""
        if (
            self.analyzer is None
            or not self.anonymization_rule_list
            or self.analyzer.presidio_analyzer_engine is None
        ):
            return text

        if not self.anonymization_rule_list.rule_list:
            return text

        analyzer_results = self.analyzer.presidio_analyzer_engine.analyze(
            text=text,
            entities=[
                rule.entity_name for rule in self.anonymization_rule_list.rule_list
            ],
            language=self.anonymization_rule_list.language,
        )

        operators = self.get_operators()

        anonymized_text = self.anonymizer_engine.anonymize(
            text=text,
            analyzer_results=analyzer_results,
            operators=operators,
        )

        return anonymized_text.text

    @staticmethod
    @typing.no_type_check  # faker is not typed correctly
    def _get_supported_faker_entities() -> Dict[Text, Any]:
        faker = Faker(["en_US", "es_ES", "it_IT"])

        # Presidio entities: https://microsoft.github.io/presidio/supported_entities/
        # Faker providers: https://faker.readthedocs.io/en/master/providers.html
        # Unsupported entities by faker:
        # CRYPTO, NRP, MEDICAL_LICENSE, US_BANK_NUMBER, US_DRIVER_LICENSE
        # UK_NHS, IT_FISCAL_CODE, IT_DRIVER_LICENSE, IT_PASSPORT, IT_IDENTITY_CARD
        # SG_NRIC_FIN, AU_ABN, AU_ACN, AU_TFN, AU_MEDICARE
        supported_entities = {
            "PERSON": lambda value: faker["en_US"].name(),
            "PHONE_NUMBER": lambda value: faker["en_US"].phone_number(),
            "EMAIL_ADDRESS": lambda value: faker["en_US"].ascii_email(),
            "CREDIT_CARD": lambda value: faker["en_US"].credit_card_number(),
            "IBAN_CODE": lambda value: faker["en_US"].iban(),
            "DATE_TIME": lambda value: faker["en_US"].date(),
            "IP_ADDRESS": lambda value: faker["en_US"].ipv4(),
            "URL": lambda value: faker["en_US"].url(),
            # This faker method returns a tuple of
            # (latitude, longitude, place name, country code, timezone)
            "LOCATION": lambda value: faker["en_US"].location_on_land()[2],
            # USA
            "US_ITIN": lambda value: faker["en_US"].itin(),
            "US_PASSPORT": lambda value: faker["en_US"].passport_number(),
            "US_SSN": lambda value: faker["en_US"].ssn(),
            # Spain
            "ES_NIF": lambda value: faker["es_ES"].nif(),
            # Italy
            "IT_VAT_CODE": lambda value: faker["it_IT"].company_vat(),
        }

        return supported_entities

    @staticmethod
    def _mask_anonymize(value: Text) -> Text:
        return "*" * len(value)

    def get_substitution_func(self, rule: AnonymizationRule) -> Optional[Any]:
        """Returns a function that anonymizes the given text.

        Args:
            rule: The anonymization rule to use.

        Returns:
            A function that anonymizes the given text.
        """
        if rule.substitution == "faker":
            supported_faker_entities = self._get_supported_faker_entities()

            if rule.entity_name not in supported_faker_entities:
                logger.debug(
                    f"Unsupported faker entity: {rule.entity_name}. "
                    f"Supported entities are: {supported_faker_entities.keys()}"
                    f"Using mask anonymization instead."
                )
                func = self._mask_anonymize
            else:
                func = supported_faker_entities.get(rule.entity_name)
        elif rule.substitution == "mask":
            func = self._mask_anonymize
        else:
            raise RasaException(f"Unknown substitution type: {rule.substitution}")

        return func

    def get_operators(self) -> Dict[Text, OperatorConfig]:
        """Returns a dictionary of operators for the given anonymization rule list."""
        operators = {}

        for rule in self.anonymization_rule_list.rule_list:
            if rule.substitution == "text":
                operators[rule.entity_name] = OperatorConfig(
                    "replace", {"new_value": rule.value}
                )
            else:
                operators[rule.entity_name] = OperatorConfig(
                    "custom", {"lambda": self.get_substitution_func(rule)}
                )

        return operators



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict
from rasa.dialogue_understanding.stack.frames import DialogueStackFrame


@dataclass
class ChitChatStackFrame(DialogueStackFrame):
    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return "chitchat"

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> ChitChatStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return ChitChatStackFrame(
            frame_id=data["frame_id"],
        )



from pathlib import Path
from typing import Tuple, Type

from rasa.engine.graph import ExecutionContext
from rasa.engine.runner.interface import GraphRunner
from rasa.engine.storage.storage import ModelMetadata, ModelStorage


def load_predict_graph_runner(
    storage_path: Path,
    model_archive_path: Path,
    model_storage_class: Type[ModelStorage],
    graph_runner_class: Type[GraphRunner],
) -> Tuple[ModelMetadata, GraphRunner]:
    """Loads a model from an archive and creates the prediction graph runner.

    Args:
        storage_path: Directory which contains the persisted graph components.
        model_archive_path: The path to the model archive.
        model_storage_class: The class to instantiate the model storage from.
        graph_runner_class: The class to instantiate the runner from.

    Returns:
        A tuple containing the model metadata and the prediction graph runner.
    """
    model_storage, model_metadata = model_storage_class.from_model_archive(
        storage_path=storage_path, model_archive_path=model_archive_path
    )
    runner = graph_runner_class.create(
        graph_schema=model_metadata.predict_schema,
        model_storage=model_storage,
        execution_context=ExecutionContext(
            graph_schema=model_metadata.predict_schema, model_id=model_metadata.model_id
        ),
    )
    return model_metadata, runner



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Text

import structlog

from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.patterns.internal_error import (
    InternalErrorPatternFlowStackFrame,
)
from rasa.shared.constants import RASA_PATTERN_INTERNAL_ERROR_DEFAULT
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class ErrorCommand(Command):
    """A command to indicate that the bot failed to handle the dialogue."""

    error_type: Text = RASA_PATTERN_INTERNAL_ERROR_DEFAULT
    info: Dict[Text, Any] = field(default_factory=dict)

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "error"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ErrorCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return ErrorCommand(
            error_type=data.get("error_type", RASA_PATTERN_INTERNAL_ERROR_DEFAULT),
            info=data.get("info", {}),
        )

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """
        stack = tracker.stack
        structlogger.debug("command_executor.error", command=self)
        stack.push(
            InternalErrorPatternFlowStackFrame(
                error_type=self.error_type, info=self.info
            )
        )
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        hashed = hash(self.error_type)
        if self.info:
            hashed += hash(str(self.info))
        return hashed

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ErrorCommand):
            return False

        return other.error_type == self.error_type and other.info == self.info



from typing import Dict, Type

from rasa.dialogue_understanding.commands import (
    CancelFlowCommand,
    CannotHandleCommand,
    ChitChatAnswerCommand,
    Command,
    HumanHandoffCommand,
    KnowledgeAnswerCommand,
    SessionStartCommand,
    SkipQuestionCommand,
    RestartCommand,
)
from rasa.dialogue_understanding.patterns.cancel import CancelPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.cannot_handle import (
    CannotHandlePatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.chitchat import ChitchatPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.human_handoff import (
    HumanHandoffPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.restart import RestartPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.search import SearchPatternFlowStackFrame
from rasa.dialogue_understanding.patterns.session_start import (
    SessionStartPatternFlowStackFrame,
)
from rasa.dialogue_understanding.patterns.skip_question import (
    SkipQuestionPatternFlowStackFrame,
)

triggerable_pattern_to_command_class: Dict[str, Type[Command]] = {
    SessionStartPatternFlowStackFrame.flow_id: SessionStartCommand,
    CancelPatternFlowStackFrame.flow_id: CancelFlowCommand,
    ChitchatPatternFlowStackFrame.flow_id: ChitChatAnswerCommand,
    HumanHandoffPatternFlowStackFrame.flow_id: HumanHandoffCommand,
    SearchPatternFlowStackFrame.flow_id: KnowledgeAnswerCommand,
    SkipQuestionPatternFlowStackFrame.flow_id: SkipQuestionCommand,
    CannotHandlePatternFlowStackFrame.flow_id: CannotHandleCommand,
    RestartPatternFlowStackFrame.flow_id: RestartCommand,
}



from __future__ import annotations
import time
import json
import logging
import os
import requests
from typing import Any, List, Optional, Text, Dict

import rasa.utils.endpoints as endpoints_utils
from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import DOCS_URL_COMPONENTS
from rasa.shared.nlu.constants import ENTITIES, TEXT
from rasa.nlu.extractors.extractor import EntityExtractorMixin
from rasa.shared.nlu.training_data.message import Message
import rasa.shared.utils.io


logger = logging.getLogger(__name__)


def extract_value(match: Dict[Text, Any]) -> Dict[Text, Any]:
    if match["value"].get("type") == "interval":
        value = {
            "to": match["value"].get("to", {}).get("value"),
            "from": match["value"].get("from", {}).get("value"),
        }
    else:
        value = match["value"].get("value")

    return value


def convert_duckling_format_to_rasa(
    matches: List[Dict[Text, Any]],
) -> List[Dict[Text, Any]]:
    extracted = []

    for match in matches:
        value = extract_value(match)
        entity = {
            "start": match["start"],
            "end": match["end"],
            "text": match.get("body", match.get("text", None)),
            "value": value,
            "confidence": 1.0,
            "additional_info": match["value"],
            "entity": match["dim"],
        }

        extracted.append(entity)

    return extracted


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.ENTITY_EXTRACTOR, is_trainable=False
)
class DucklingEntityExtractor(GraphComponent, EntityExtractorMixin):
    """Searches for structured entities, e.g. dates, using a duckling server."""

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """The component's default config (see parent class for full docstring)."""
        return {
            # by default all dimensions recognized by duckling are returned
            # dimensions can be configured to contain an array of strings
            # with the names of the dimensions to filter for
            "dimensions": None,
            # http url of the running duckling server
            "url": None,
            # locale - if not set, we will use the language of the model
            "locale": None,
            # timezone like Europe/Berlin
            # if not set the default timezone of Duckling is going to be used
            "timezone": None,
            # Timeout for receiving response from HTTP URL of the running
            # duckling server. If not set the default timeout of duckling HTTP URL
            # is set to 3 seconds.
            "timeout": 3,
        }

    def __init__(self, config: Dict[Text, Any]) -> None:
        """Creates the extractor.

        Args:
            config: The extractor's config.
        """
        self.component_config = config

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> DucklingEntityExtractor:
        """Creates component (see parent class for full docstring)."""
        return cls(config)

    def _url(self) -> Optional[Text]:
        """Return url of the duckling service. Environment var will override."""
        if os.environ.get("RASA_DUCKLING_HTTP_URL"):
            return os.environ["RASA_DUCKLING_HTTP_URL"]

        return self.component_config.get("url")

    def _payload(self, text: Text, reference_time: int) -> Dict[Text, Any]:
        dimensions = self.component_config["dimensions"]
        return {
            "text": text,
            "locale": self.component_config["locale"],
            "tz": self.component_config.get("timezone"),
            "dims": json.dumps(dimensions),
            "reftime": reference_time,
        }

    def _duckling_parse(self, text: Text, reference_time: int) -> List[Dict[Text, Any]]:
        """Sends the request to the duckling server and parses the result.

        Args:
            text: Text for duckling server to parse.
            reference_time: Reference time in milliseconds.

        Returns:
            JSON response from duckling server with parse data.
        """
        parse_url = endpoints_utils.concat_url(self._url(), "/parse")
        try:
            payload = self._payload(text, reference_time)
            headers = {
                "Content-Type": "application/x-www-form-urlencoded; charset=UTF-8"
            }
            response = requests.post(
                parse_url,
                data=payload,
                headers=headers,
                timeout=self.component_config.get("timeout"),
            )
            if response.status_code == 200:
                return response.json()
            else:
                logger.error(
                    f"Failed to get a proper response from remote "
                    f"duckling at '{parse_url}. "
                    f"Status Code: {response.status_code}. "
                    f"Response: {response.text}"
                )
                return []
        except (
            requests.exceptions.ConnectionError,
            requests.exceptions.ReadTimeout,
        ) as e:
            logger.error(
                "Failed to connect to duckling http server. Make sure "
                "the duckling server is running/healthy/not stale and the proper host "
                "and port are set in the configuration. More "
                "information on how to run the server can be found on "
                "github: "
                "https://github.com/facebook/duckling#quickstart "
                "Error: {}".format(e)
            )
            return []

    @staticmethod
    def _reference_time_from_message(message: Message) -> int:
        if message.time is not None:
            try:
                return message.time * 1000
            except ValueError as e:
                logging.warning(
                    "Could not parse timestamp {}. Instead "
                    "current UTC time will be passed to "
                    "duckling. Error: {}".format(message.time, e)
                )
        # fallbacks to current time, multiplied by 1000 because duckling
        # requires the reftime in milliseconds
        return int(time.time()) * 1000

    def process(self, messages: List[Message]) -> List[Message]:
        """Augments the message with potentially extracted entities."""
        if self._url() is None:
            rasa.shared.utils.io.raise_warning(
                "Duckling HTTP component in pipeline, but no "
                "`url` configuration in the config "
                "file nor is `RASA_DUCKLING_HTTP_URL` "
                "set as an environment variable. No entities will be extracted!",
                docs=DOCS_URL_COMPONENTS + "#DucklingEntityExtractor",
            )
            return messages

        for message in messages:
            reference_time = self._reference_time_from_message(message)
            matches = self._duckling_parse(message.get(TEXT), reference_time)
            all_extracted = convert_duckling_format_to_rasa(matches)
            dimensions = self.component_config["dimensions"]
            extracted = self.filter_irrelevant_entities(all_extracted, dimensions)
            extracted = self.add_extractor_name(extracted)
            message.set(
                ENTITIES, message.get(ENTITIES, []) + extracted, add_to_output=True
            )

        return messages



from __future__ import annotations
from dataclasses import dataclass
from typing import Any, Dict

from rasa.dialogue_understanding.stack.frames import DialogueStackFrame


@dataclass
class SearchStackFrame(DialogueStackFrame):
    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return "search"

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> SearchStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        return SearchStackFrame(
            frame_id=data["frame_id"],
        )



import os
import re
from typing import Dict, Any, Optional

import structlog

from rasa.shared.constants import (
    OPENAI_API_BASE_ENV_VAR,
    OPENAI_API_VERSION_ENV_VAR,
    AZURE_API_BASE_ENV_VAR,
    AZURE_API_VERSION_ENV_VAR,
    API_BASE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    DEPLOYMENT_CONFIG_KEY,
    AZURE_API_KEY_ENV_VAR,
    OPENAI_API_TYPE_ENV_VAR,
    OPENAI_API_KEY_ENV_VAR,
    AZURE_API_TYPE_ENV_VAR,
    AZURE_OPENAI_PROVIDER,
)
from rasa.shared.exceptions import ProviderClientValidationError
from rasa.shared.providers._configs.azure_openai_client_config import (
    AzureOpenAIClientConfig,
)
from rasa.shared.providers.llm._base_litellm_client import _BaseLiteLLMClient
from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()


class AzureOpenAILLMClient(_BaseLiteLLMClient):
    """
    A client for interfacing with Azure's OpenAI LLM deployments.

    Parameters:
        deployment (str): The deployment name.
        model (Optional[str]): The name of the deployed model.
        api_type: (Optional[str]): The api type. If not provided, it will be set via
            environment variable.
        api_base (Optional[str]): The base URL for the API endpoints. If not provided,
            it will be set via environment variables.
        api_version (Optional[str]): The version of the API to use. If not provided,
            it will be set via environment variable.
        kwargs (Optional[Dict[str, Any]]): Optional configuration parameters specific
            to the model deployment.

    Raises:
        ProviderClientValidationError: If validation of the client setup fails.
        DeprecationWarning: If deprecated environment variables are used for
            configuration.
    """

    def __init__(
        self,
        deployment: str,
        model: Optional[str] = None,
        api_type: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__()  # type: ignore
        self._deployment = deployment
        self._model = model
        self._extra_parameters = kwargs or {}

        # Set api_base with the following priority:
        # parameter -> Azure Env Var -> (deprecated) OpenAI Env Var
        self._api_base = (
            api_base
            or os.getenv(AZURE_API_BASE_ENV_VAR)
            or os.getenv(OPENAI_API_BASE_ENV_VAR)
        )

        # Set api_version with the following priority:
        # parameter -> Azure Env Var -> (deprecated) OpenAI Env Var
        self._api_version = (
            api_version
            or os.getenv(AZURE_API_VERSION_ENV_VAR)
            or os.getenv(OPENAI_API_VERSION_ENV_VAR)
        )

        # API key can be set through OPENAI_API_KEY too,
        # because of the backward compatibility
        self._api_key = os.getenv(AZURE_API_KEY_ENV_VAR) or os.getenv(
            OPENAI_API_KEY_ENV_VAR
        )

        # Not used by LiteLLM, here for backward compatibility
        self._api_type = (
            api_type
            or os.getenv(AZURE_API_TYPE_ENV_VAR)
            or os.getenv(OPENAI_API_TYPE_ENV_VAR)
        )

        # Run helper function to check and raise deprecation warning if
        # deprecated environment variables were used for initialization of the
        # client settings
        self._raise_evn_var_deprecation_warnings()

        # validate the client settings
        self.validate_client_setup()

    def _raise_evn_var_deprecation_warnings(self) -> None:
        """Helper function to check and raise deprecation warning if
        deprecated environment variables were used for initialization of
        some client settings.
        """
        deprecation_mapping = {
            "API Base": {
                "current_value": self.api_base,
                "env_var": AZURE_API_BASE_ENV_VAR,
                "deprecated_var": OPENAI_API_BASE_ENV_VAR,
            },
            "API Version": {
                "current_value": self.api_version,
                "env_var": AZURE_API_VERSION_ENV_VAR,
                "deprecated_var": OPENAI_API_VERSION_ENV_VAR,
            },
            "API Key": {
                "current_value": self._api_key,
                "env_var": AZURE_API_KEY_ENV_VAR,
                "deprecated_var": OPENAI_API_KEY_ENV_VAR,
            },
        }

        deprecation_warning_message = (
            "Usage of {deprecated_env_var} environment "
            "variable for setting the {setting} for Azure "
            "OpenAI client is deprecated and will be removed "
            "in 4.0.0. "
        )
        deprecation_warning_replacement_message = (
            "Please use {env_var} environment variable."
        )

        for setting in deprecation_mapping.keys():
            current_value = deprecation_mapping[setting]["current_value"]
            env_var = deprecation_mapping[setting]["env_var"]
            deprecated_var = deprecation_mapping[setting]["deprecated_var"]

            # Value is set through the non-deprecated env var
            if current_value == os.getenv(env_var):
                continue

            # Value is set through the deprecated env var
            if current_value == os.getenv(deprecated_var):
                message = deprecation_warning_message.format(
                    setting=setting, deprecated_env_var=deprecated_var
                )
                if env_var is not None:
                    message += deprecation_warning_replacement_message.format(
                        env_var=env_var
                    )
                raise_deprecation_warning(message=message)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "AzureOpenAILLMClient":
        """
        Initializes the client from given configuration.

        Args:
            config (Dict[str, Any]): Configuration.

        Raises:
            ValueError:
                If any of the required configuration keys are missing.
                If `api_type` has a value different from `azure`.

        Returns:
            AzureOpenAILLMClient: Initialized client.
        """
        try:
            azure_openai_config = AzureOpenAIClientConfig.from_dict(config)
        except ValueError as e:
            message = "Cannot instantiate a client from the passed configuration."
            structlogger.error(
                "azure_openai_llm_client.from_config.error",
                message=message,
                config=config,
                original_error=e,
            )
            raise

        return cls(
            azure_openai_config.deployment,
            azure_openai_config.model,
            azure_openai_config.api_type,
            azure_openai_config.api_base,
            azure_openai_config.api_version,
            **azure_openai_config.extra_parameters,
        )

    @property
    def config(self) -> dict:
        """Returns the configuration for that the llm client
        in dictionary form.
        """
        config = AzureOpenAIClientConfig(
            deployment=self._deployment,
            model=self._model,
            api_base=self._api_base,
            api_version=self._api_version,
            api_type=self._api_type,
            extra_parameters=self._extra_parameters,
        )
        return config.to_dict()

    @property
    def deployment(self) -> str:
        return self._deployment

    @property
    def model(self) -> Optional[str]:
        """
        Returns the name of the model deployed on Azure.
        """
        return self._model

    @property
    def api_base(self) -> Optional[str]:
        """
        Returns the API base URL for the Azure OpenAI llm client.
        """
        return self._api_base

    @property
    def api_version(self) -> Optional[str]:
        """
        Returns the API version for the Azure OpenAI llm client.
        """
        return self._api_version

    @property
    def api_type(self) -> Optional[str]:
        return self._api_type

    @property
    def _litellm_model_name(self) -> str:
        """Returns the value of LiteLLM's model parameter to be used in
        completion/acompletion in LiteLLM format:

        <provider>/<model or deployment name>
        """
        regex_pattern = rf"^{AZURE_OPENAI_PROVIDER}/"
        if not re.match(regex_pattern, self._deployment):
            return f"{AZURE_OPENAI_PROVIDER}/{self._deployment}"
        return self._deployment

    @property
    def _litellm_extra_parameters(self) -> Dict[str, Any]:
        return self._extra_parameters

    @property
    def _completion_fn_args(self) -> Dict[str, Any]:
        """Returns the completion arguments for invoking a call through
        LiteLLM's completion functions.
        """
        fn_args = super()._completion_fn_args
        fn_args.update(
            {
                "api_base": self.api_base,
                "api_version": self.api_version,
                "api_key": self._api_key,
            }
        )
        return fn_args

    def validate_client_setup(self) -> None:
        """Validates that all required configuration parameters are set."""

        def generate_event_info_for_missing_setting(
            setting: str,
            setting_env_var: Optional[str] = None,
            setting_config_key: Optional[str] = None,
        ) -> str:
            """Generate a part of the message with instructions on what to set
            for the missing client setting.
            """
            info = "Set {setting} with {options}. "
            options = ""
            if setting_env_var is not None:
                options += f"environment variable '{setting_env_var}'"
            if setting_config_key is not None and setting_env_var is not None:
                options += " or "
            if setting_config_key is not None:
                options += f"config key '{setting_config_key}'"

            return info.format(setting=setting, options=options)

        # All required settings for Azure OpenAI client
        settings: Dict[str, Dict[str, Any]] = {
            "API Base": {
                "current_value": self.api_base,
                "env_var": AZURE_API_BASE_ENV_VAR,
                "config_key": API_BASE_CONFIG_KEY,
            },
            "API Version": {
                "current_value": self.api_version,
                "env_var": AZURE_API_VERSION_ENV_VAR,
                "config_key": API_VERSION_CONFIG_KEY,
            },
            "Deployment Name": {
                "current_value": self.deployment,
                "env_var": None,
                "config_key": DEPLOYMENT_CONFIG_KEY,
            },
            "API Key": {
                "current_value": self._api_key,
                "env_var": AZURE_API_KEY_ENV_VAR,
                "config_key": None,
            },
        }

        missing_settings = [
            setting_name
            for setting_name, setting_info in settings.items()
            if setting_info["current_value"] is None
        ]

        if missing_settings:
            event_info = f"Client settings not set: " f"{', '.join(missing_settings)}. "

            for missing_setting in missing_settings:
                if settings[missing_setting]["current_value"] is not None:
                    continue
                event_info += generate_event_info_for_missing_setting(
                    missing_setting,
                    settings[missing_setting]["env_var"],
                    settings[missing_setting]["config_key"],
                )

            structlogger.error(
                "azure_openai_llm_client.not_configured",
                event_info=event_info,
                missing_settings=missing_settings,
            )
            raise ProviderClientValidationError(event_info)



from typing import Dict, Type, Optional

from rasa.shared.constants import (
    AZURE_OPENAI_PROVIDER,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER,
    OPENAI_PROVIDER,
    SELF_HOSTED_PROVIDER,
)
from rasa.shared.providers.embedding.azure_openai_embedding_client import (
    AzureOpenAIEmbeddingClient,
)
from rasa.shared.providers.embedding.default_litellm_embedding_client import (
    DefaultLiteLLMEmbeddingClient,
)
from rasa.shared.providers.embedding.embedding_client import EmbeddingClient
from rasa.shared.providers.embedding.huggingface_local_embedding_client import (
    HuggingFaceLocalEmbeddingClient,
)
from rasa.shared.providers.embedding.openai_embedding_client import (
    OpenAIEmbeddingClient,
)
from rasa.shared.providers.llm.azure_openai_llm_client import AzureOpenAILLMClient
from rasa.shared.providers.llm.default_litellm_llm_client import DefaultLiteLLMClient
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.providers.llm.openai_llm_client import OpenAILLMClient
from rasa.shared.providers.llm.self_hosted_llm_client import SelfHostedLLMClient
from rasa.shared.providers._configs.azure_openai_client_config import (
    AzureOpenAIClientConfig,
)
from rasa.shared.providers._configs.default_litellm_client_config import (
    DefaultLiteLLMClientConfig,
)
from rasa.shared.providers._configs.huggingface_local_embedding_client_config import (
    HuggingFaceLocalEmbeddingClientConfig,
)
from rasa.shared.providers._configs.openai_client_config import OpenAIClientConfig
from rasa.shared.providers._configs.self_hosted_llm_client_config import (
    SelfHostedLLMClientConfig,
)
from rasa.shared.providers._configs.client_config import ClientConfig

_provider_to_llm_client_mapping: Dict[str, Type[LLMClient]] = {
    OPENAI_PROVIDER: OpenAILLMClient,
    AZURE_OPENAI_PROVIDER: AzureOpenAILLMClient,
    SELF_HOSTED_PROVIDER: SelfHostedLLMClient,
}

_provider_to_embedding_client_mapping: Dict[str, Type[EmbeddingClient]] = {
    OPENAI_PROVIDER: OpenAIEmbeddingClient,
    AZURE_OPENAI_PROVIDER: AzureOpenAIEmbeddingClient,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER: HuggingFaceLocalEmbeddingClient,
}

_provider_to_client_config_class_mapping: Dict[str, Type] = {
    OPENAI_PROVIDER: OpenAIClientConfig,
    AZURE_OPENAI_PROVIDER: AzureOpenAIClientConfig,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER: HuggingFaceLocalEmbeddingClientConfig,
    SELF_HOSTED_PROVIDER: SelfHostedLLMClientConfig,
}


def get_llm_client_from_provider(provider: Optional[str]) -> Type[LLMClient]:
    return _provider_to_llm_client_mapping.get(provider, DefaultLiteLLMClient)


def get_embedding_client_from_provider(provider: str) -> Type[EmbeddingClient]:
    return _provider_to_embedding_client_mapping.get(
        provider, DefaultLiteLLMEmbeddingClient
    )


def get_client_config_class_from_provider(provider: str) -> Type[ClientConfig]:
    return _provider_to_client_config_class_mapping.get(
        provider, DefaultLiteLLMClientConfig
    )



from __future__ import annotations
import asyncio
from contextlib import asynccontextmanager
import json
import logging
import os

from typing import AsyncGenerator, Dict, Optional, Text, Union

from rasa.shared.exceptions import RasaException, ConnectionException
import rasa.shared.utils.common
from rasa.core.constants import DEFAULT_LOCK_LIFETIME
from rasa.core.lock import TicketLock
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


def _get_lock_lifetime() -> int:
    return int(os.environ.get("TICKET_LOCK_LIFETIME", 0)) or DEFAULT_LOCK_LIFETIME


LOCK_LIFETIME = _get_lock_lifetime()
DEFAULT_SOCKET_TIMEOUT_IN_SECONDS = 10

DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX = "lock:"


# noinspection PyUnresolvedReferences
class LockError(RasaException):
    """Exception that is raised when a lock cannot be acquired.

    Attributes:
         message (str): explanation of which `conversation_id` raised the error
    """

    pass


class LockStore:
    """Base class for ticket locks."""

    @staticmethod
    def create(obj: Union[LockStore, EndpointConfig, None]) -> LockStore:
        """Factory to create a lock store."""
        if isinstance(obj, LockStore):
            return obj

        try:
            return _create_from_endpoint_config(obj)
        except ConnectionError as error:
            raise ConnectionException("Cannot connect to lock store.") from error

    @staticmethod
    def create_lock(conversation_id: Text) -> TicketLock:
        """Create a new `TicketLock` for `conversation_id`."""
        return TicketLock(conversation_id)

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Fetch lock for `conversation_id` from storage."""
        raise NotImplementedError

    def delete_lock(self, conversation_id: Text) -> None:
        """Delete lock for `conversation_id` from storage."""
        raise NotImplementedError

    def save_lock(self, lock: TicketLock) -> None:
        """Commit `lock` to storage."""
        raise NotImplementedError

    def issue_ticket(
        self, conversation_id: Text, lock_lifetime: float = LOCK_LIFETIME
    ) -> int:
        """Issue new ticket with `lock_lifetime` for lock associated with
        `conversation_id`.

        Creates a new lock if none is found.
        """
        logger.debug(f"Issuing ticket for conversation '{conversation_id}'.")
        try:
            lock = self.get_or_create_lock(conversation_id)
            ticket = lock.issue_ticket(lock_lifetime)
            self.save_lock(lock)

            return ticket
        except Exception as e:
            raise LockError(f"Error while acquiring lock. Error:\n{e}")

    @asynccontextmanager
    async def lock(
        self,
        conversation_id: Text,
        lock_lifetime: float = LOCK_LIFETIME,
        wait_time_in_seconds: float = 1,
    ) -> AsyncGenerator[TicketLock, None]:
        """Acquire lock with lifetime `lock_lifetime`for `conversation_id`.

        Try acquiring lock with a wait time of `wait_time_in_seconds` seconds
        between attempts. Raise a `LockError` if lock has expired.
        """
        ticket = self.issue_ticket(conversation_id, lock_lifetime)
        try:
            yield await self._acquire_lock(
                conversation_id, ticket, wait_time_in_seconds
            )
        finally:
            self.cleanup(conversation_id, ticket)

    async def _acquire_lock(
        self, conversation_id: Text, ticket: int, wait_time_in_seconds: float
    ) -> TicketLock:
        logger.debug(f"Acquiring lock for conversation '{conversation_id}'.")
        while True:
            # fetch lock in every iteration because lock might no longer exist
            lock = self.get_lock(conversation_id)

            # exit loop if lock does not exist anymore (expired)
            if not lock:
                break

            # acquire lock if it isn't locked
            if not lock.is_locked(ticket):
                logger.debug(f"Acquired lock for conversation '{conversation_id}'.")
                return lock

            items_before_this = ticket - (lock.now_serving or 0)

            logger.debug(
                f"Failed to acquire lock for conversation ID '{conversation_id}' "
                f"because {items_before_this} other item(s) for this "
                f"conversation ID have to be finished processing first. "
                f"Retrying in {wait_time_in_seconds} seconds ..."
            )

            # sleep and update lock
            await asyncio.sleep(wait_time_in_seconds)
            self.update_lock(conversation_id)

        raise LockError(
            f"Could not acquire lock for conversation_id '{conversation_id}'."
        )

    def update_lock(self, conversation_id: Text) -> None:
        """Fetch lock for `conversation_id`, remove expired tickets and save lock."""
        lock = self.get_lock(conversation_id)
        if lock:
            lock.remove_expired_tickets()
            self.save_lock(lock)

    def get_or_create_lock(self, conversation_id: Text) -> TicketLock:
        """Fetch existing lock for `conversation_id`.

        Alternatively, create a new one if it doesn't exist.
        """
        existing_lock = self.get_lock(conversation_id)

        if existing_lock:
            return existing_lock

        return self.create_lock(conversation_id)

    def is_someone_waiting(self, conversation_id: Text) -> bool:
        """Return whether someone is waiting for lock for this `conversation_id`."""
        lock = self.get_lock(conversation_id)
        if lock:
            return lock.is_someone_waiting()

        return False

    def finish_serving(self, conversation_id: Text, ticket_number: int) -> None:
        """Finish serving ticket with `ticket_number` for `conversation_id`.

        Removes ticket from lock and saves lock.
        """
        lock = self.get_lock(conversation_id)
        if lock:
            lock.remove_ticket_for(ticket_number)
            self.save_lock(lock)

    def cleanup(self, conversation_id: Text, ticket_number: int) -> None:
        """Remove lock for `conversation_id` if no one is waiting."""
        self.finish_serving(conversation_id, ticket_number)
        if not self.is_someone_waiting(conversation_id):
            self.delete_lock(conversation_id)

    @staticmethod
    def _log_deletion(conversation_id: Text, deletion_successful: bool) -> None:
        if deletion_successful:
            logger.debug(f"Deleted lock for conversation '{conversation_id}'.")
        else:
            logger.debug(f"Could not delete lock for conversation '{conversation_id}'.")


class RedisLockStore(LockStore):
    """Redis store for ticket locks."""

    def __init__(
        self,
        host: Text = "localhost",
        port: int = 6379,
        db: int = 1,
        username: Optional[Text] = None,
        password: Optional[Text] = None,
        use_ssl: bool = False,
        ssl_certfile: Optional[Text] = None,
        ssl_keyfile: Optional[Text] = None,
        ssl_ca_certs: Optional[Text] = None,
        key_prefix: Optional[Text] = None,
        socket_timeout: float = DEFAULT_SOCKET_TIMEOUT_IN_SECONDS,
    ) -> None:
        """Create a lock store which uses Redis for persistence.

        Args:
            host: The host of the redis server.
            port: The port of the redis server.
            db: The name of the database within Redis which should be used by Rasa
                Open Source.
            username: The username which should be used for authentication with the
                Redis database.
            password: The password which should be used for authentication with the
                Redis database.
            use_ssl: `True` if SSL should be used for the connection to Redis.
            ssl_certfile: Path to the SSL certificate file.
            ssl_keyfile: Path to the SSL private key file.
            ssl_ca_certs: Path to the SSL CA certificate file.
            key_prefix: prefix to prepend to all keys used by the lock store. Must be
                alphanumeric.
            socket_timeout: Timeout in seconds after which an exception will be raised
                in case Redis doesn't respond within `socket_timeout` seconds.
        """
        import redis

        self.red = redis.StrictRedis(
            host=host,
            port=int(port),
            db=int(db),
            username=username,
            password=password,
            ssl=use_ssl,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            ssl_ca_certs=ssl_ca_certs,
            socket_timeout=socket_timeout,
        )

        self.key_prefix = DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX
        if key_prefix:
            logger.debug(f"Setting non-default redis key prefix: '{key_prefix}'.")
            self._set_key_prefix(key_prefix)

        super().__init__()

    def _set_key_prefix(self, key_prefix: Text) -> None:
        if isinstance(key_prefix, str) and key_prefix.isalnum():
            self.key_prefix = key_prefix + ":" + DEFAULT_REDIS_LOCK_STORE_KEY_PREFIX
        else:
            logger.warning(
                f"Omitting provided non-alphanumeric redis key prefix: '{key_prefix}'. "
                f"Using default '{self.key_prefix}' instead."
            )

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Retrieves lock (see parent docstring for more information)."""
        serialised_lock = self.red.get(self.key_prefix + conversation_id)
        if serialised_lock:
            return TicketLock.from_dict(json.loads(serialised_lock))

        return None

    def delete_lock(self, conversation_id: Text) -> None:
        """Deletes lock for conversation ID."""
        deletion_successful = self.red.delete(self.key_prefix + conversation_id)
        self._log_deletion(conversation_id, deletion_successful)

    def save_lock(self, lock: TicketLock) -> None:
        self.red.set(self.key_prefix + lock.conversation_id, lock.dumps())


class InMemoryLockStore(LockStore):
    """In-memory store for ticket locks."""

    def __init__(self) -> None:
        """Initialise dictionary of locks."""
        self.conversation_locks: Dict[Text, TicketLock] = {}
        super().__init__()

    def get_lock(self, conversation_id: Text) -> Optional[TicketLock]:
        """Get lock for conversation if it exists."""
        return self.conversation_locks.get(conversation_id)

    def delete_lock(self, conversation_id: Text) -> None:
        """Delete lock for conversation."""
        deleted_lock = self.conversation_locks.pop(conversation_id, None)
        self._log_deletion(
            conversation_id, deletion_successful=deleted_lock is not None
        )

    def save_lock(self, lock: TicketLock) -> None:
        """Save lock in store."""
        self.conversation_locks[lock.conversation_id] = lock


def _create_from_endpoint_config(
    endpoint_config: Optional[EndpointConfig] = None,
) -> LockStore:
    """Given an endpoint configuration, create a proper `LockStore` object."""
    if (
        endpoint_config is None
        or endpoint_config.type is None
        or endpoint_config.type == "in_memory"
    ):
        # this is the default type if no lock store type is set

        lock_store: LockStore = InMemoryLockStore()
    elif endpoint_config.type == "redis":
        lock_store = RedisLockStore(host=endpoint_config.url, **endpoint_config.kwargs)
    elif endpoint_config.type == "concurrent_redis":
        from rasa.core.concurrent_lock_store import ConcurrentRedisLockStore

        lock_store = ConcurrentRedisLockStore(endpoint_config=endpoint_config)
    else:
        lock_store = _load_from_module_name_in_endpoint_config(endpoint_config)

    logger.debug(f"Connected to lock store '{lock_store.__class__.__name__}'.")

    return lock_store


def _load_from_module_name_in_endpoint_config(
    endpoint_config: EndpointConfig,
) -> LockStore:
    """Retrieve a `LockStore` based on its class name."""
    try:
        lock_store_class = rasa.shared.utils.common.class_from_module_path(
            endpoint_config.type
        )
        return lock_store_class(endpoint_config=endpoint_config)
    except (AttributeError, ImportError) as e:
        raise Exception(
            f"Could not find a class based on the module path "
            f"'{endpoint_config.type}'. Failed to create a `LockStore` "
            f"instance. Error: {e}"
        )



import logging

from rasa.engine.recipes.recipe import Recipe
from rasa.engine.graph import GraphModelConfiguration
from rasa.shared.constants import DOCS_URL_GRAPH_RECIPE, ASSISTANT_ID_KEY
from rasa.shared.data import TrainingType
from rasa.shared.exceptions import InvalidConfigException
from rasa.shared.utils.common import mark_as_experimental_feature
from rasa.shared.utils.io import raise_warning
from rasa.engine.graph import GraphSchema

from typing import Dict, Text, Any, Tuple


logger = logging.getLogger(__name__)


class GraphV1Recipe(Recipe):
    """Recipe which converts the graph model config to train and predict graph."""

    name = "graph.v1"

    def get_targets(
        self, config: Dict, training_type: TrainingType
    ) -> Tuple[Text, Any]:
        """Return NLU and core targets from config dictionary.

        Note that default recipe has `nlu_target` and `core_target` as
        fixed values of `run_RegexMessageHandler` and `select_prediction`
        respectively. For graph recipe, target values are customizable. These
        can be used in validation (default recipe does this validation check)
        and during execution (all recipes use targets during execution).
        """
        if training_type == TrainingType.NLU:
            core_required = False
            core_target = None
        else:
            core_required = True
            core_target = config.get("core_target")
        # NLU target is required because core (prediction) depends on NLU.
        nlu_target = config.get("nlu_target")
        if nlu_target is None or (core_required and core_target is None):
            raise InvalidConfigException(
                "Can't find target names for NLU and/or core. Please make "
                "sure to provide 'nlu_target' (required for all training types) "
                "and 'core_target' (required if training is not just NLU) values in "
                "your config.yml file."
            )
        return nlu_target, core_target

    def graph_config_for_recipe(
        self,
        config: Dict,
        cli_parameters: Dict[Text, Any],
        training_type: TrainingType = TrainingType.BOTH,
        is_finetuning: bool = False,
    ) -> GraphModelConfiguration:
        """Converts the default config to graphs (see interface for full docstring)."""
        mark_as_experimental_feature("graph recipe")
        if cli_parameters or is_finetuning:
            raise_warning(
                "Unlike the Default Recipe, Graph Recipe does not utilize CLI "
                "parameters or finetuning and these configurations will be ignored. "
                "Add configuration to the recipe itself if you want them to be used.",
                docs=DOCS_URL_GRAPH_RECIPE,
            )

        nlu_target, core_target = self.get_targets(config, training_type)

        return GraphModelConfiguration(
            train_schema=GraphSchema.from_dict(config.get("train_schema")),
            predict_schema=GraphSchema.from_dict(config.get("predict_schema")),
            training_type=training_type,
            assistant_id=config.get(ASSISTANT_ID_KEY),
            language=config.get("language"),
            spaces=config.get("spaces"),
            core_target=core_target,
            nlu_target=nlu_target,
        )



from typing import Protocol, Dict, List, TYPE_CHECKING, runtime_checkable

if TYPE_CHECKING:
    from rasa.shared.providers.embedding.embedding_response import EmbeddingResponse


@runtime_checkable
class EmbeddingClient(Protocol):
    @classmethod
    def from_config(cls, config: dict) -> "EmbeddingClient":
        """
        Initializes the embedding client with the given configuration.

        This class method should be implemented to parse the given
        configuration and create an instance of llm client.
        """
        ...

    @property
    def config(self) -> Dict:
        """
        Returns the configuration for the embedding client.

        This property should be implemented to return a dictionary containing
        the configuration settings for the embedding client.
        """
        ...

    def embed(self, documents: List[str]) -> "EmbeddingResponse":
        """
        Embeds a list of documents synchronously.

        This method should be implemented to take a list of documents (as strings)
        and return a list of embedding vectors (each represented as a list of floats).

        Args:
            documents: List of documents to be embedded.

        Returns:
            List of embedding vectors.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        ...

    async def aembed(self, documents: List[str]) -> "EmbeddingResponse":
        """
        Embeds a list of documents asynchronously.

        This method should be implemented to take a list of documents (as strings)
        and return a list of embedding vectors (each represented as a list of floats).

        Args:
            documents: List of documents to be embedded.

        Returns:
            List of embedding vectors.

        Raises:
            ProviderClientAPIException: If API calls raised an error.
        """
        ...

    def validate_documents(self, documents: List[str]) -> None:
        """
        Validates a list of documents to ensure they are suitable for embedding.

        This method should be implemented for document validation. It should raise
        a ValueError if any document is invalid.

        Args:
            documents: List of documents to be validated.

        Raises:
            ValueError: If any document is invalid.
        """
        ...

    def validate_client_setup(self, *args, **kwargs) -> None:  # type: ignore
        """
        Perform client setup validation.

        This method should be implemented to validate whether the client can be used
        with the parameters provided through configuration or environment variables.

        If there are any issues, the client should raise a ValidationError.
        If no validation is needed, this check can simply pass.
        """
        ...



from rasa.core.information_retrieval.information_retrieval import (  # noqa: F401
    InformationRetrieval,
    SearchResult,
    SearchResultList,
    InformationRetrievalException,
    create_from_endpoint_config,
)



from __future__ import annotations
import os
import logging
import sys
from typing import Any, Dict, Optional

import structlog
from structlog_sentry import SentryProcessor
from structlog.dev import ConsoleRenderer
from structlog.typing import EventDict, WrappedLogger
from rasa.shared.constants import (
    ENV_LOG_LEVEL,
    DEFAULT_LOG_LEVEL,
    ENV_LOG_LEVEL_LLM,
    ENV_LOG_LEVEL_LLM_MODULE_NAMES,
    DEFAULT_LOG_LEVEL_LLM,
)
from rasa.plugin import plugin_manager


FORCE_JSON_LOGGING = os.environ.get("FORCE_JSON_LOGGING")


class HumanConsoleRenderer(ConsoleRenderer):
    """Console renderer that outputs human-readable logs."""

    def __call__(self, logger: WrappedLogger, name: str, event_dict: EventDict) -> str:
        if "event_info" in event_dict:
            event_key = event_dict["event"]
            event_dict["event"] = event_dict["event_info"]
            event_dict["event_key"] = event_key
            del event_dict["event_info"]

        return super().__call__(logger, name, event_dict)


def _anonymizer(
    _: structlog.BoundLogger, __: str, event_dict: Dict[str, Any]
) -> Dict[str, Any]:
    """Anonymizes event dict."""
    anonymizable_keys = [
        "text",
        "response_text",
        "user_text",
        "slots",
        "parse_data_text",
        "parse_data_entities",
        "prediction_events",
        "tracker_latest_message",
        "prefilled_slots",
        "message",
        "response",
        "slot_candidates",
        "rasa_event",
        "rasa_events",
        "tracker_states",
        "current_states",
        "old_states",
        "current_states",
        "successes",
        "current_entity",
        "next_entity",
        "states",
        "entity",
        "token_text",
        "user_message",
        "json_message",
    ]
    anonymization_pipeline = plugin_manager().hook.get_anonymization_pipeline()

    if anonymization_pipeline:
        for key in anonymizable_keys:
            if key in event_dict:
                anonymized_value = anonymization_pipeline.log_run(event_dict[key])
                event_dict[key] = anonymized_value
    return event_dict


def configure_structlog(
    log_level: Optional[int] = None,
) -> None:
    """Configure logging of the server."""
    if log_level is None:  # Log level NOTSET is 0 so we use `is None` here
        log_level_name = os.environ.get(ENV_LOG_LEVEL, DEFAULT_LOG_LEVEL)
        # Change log level from str to int (note that log_level in function parameter
        # int already, coming from CLI argparse parameter).
        log_level = logging.getLevelName(log_level_name)

    logging.basicConfig(
        format="%(message)s",
        stream=sys.stdout,
        level=log_level,
    )

    shared_processors = [
        _anonymizer,
        # Processors that have nothing to do with output,
        # e.g., add timestamps or log level names.
        # If log level is too low, abort pipeline and throw away log entry.
        structlog.stdlib.filter_by_level,
        structlog.contextvars.merge_contextvars,
        # Add the name of the logger to event dict.
        # structlog.stdlib.add_logger_name,
        # Add log level to event dict.
        structlog.processors.add_log_level,
        # If the "stack_info" key in the event dict is true, remove it and
        # render the current stack trace in the "stack" key.
        structlog.processors.StackInfoRenderer(),
        # If some value is in bytes, decode it to a unicode str.
        structlog.processors.UnicodeDecoder(),
        structlog.dev.set_exc_info,
        # add structlog sentry integration. only log fatal log entries
        # as events as we are tracking exceptions anyways
        SentryProcessor(event_level=logging.FATAL),
    ]

    if not FORCE_JSON_LOGGING and sys.stderr.isatty():
        # Pretty printing when we run in a terminal session.
        # Automatically prints pretty tracebacks when "rich" is installed
        processors = shared_processors + [
            HumanConsoleRenderer(),
        ]
    else:
        # Print JSON when we run, e.g., in a Docker container.
        # Also print structured tracebacks.
        processors = shared_processors + [
            structlog.processors.dict_tracebacks,
            structlog.processors.JSONRenderer(),
        ]

    structlog.configure(
        processors=processors,  # type: ignore
        context_class=dict,
        # `logger_factory` is used to create wrapped loggers that are used for
        # OUTPUT. This one returns a `logging.Logger`. The final value (a JSON
        # string) from the final processor (`JSONRenderer`) will be passed to
        # the method of the same name as that you've called on the bound logger.
        logger_factory=structlog.stdlib.LoggerFactory(),
        # `wrapper_class` is the bound logger that you get back from
        # get_logger(). This one imitates the API of `logging.Logger`.
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
        # Effectively freeze configuration after creating the first bound
        # logger.
        cache_logger_on_first_use=True,
    )


def log_llm(logger: Any, log_module: str, log_event: str, **kwargs: Any) -> None:
    """Logs LLM-specific events depending on a flag passed through an environment
    variable. If the module's flag is set to INFO (e.g.
    LOG_PROMPT_LLM_COMMAND_GENERATOR=INFO), its prompt is logged at INFO level,
    overriding the general log level setting.

    Args:
        logger: instance of the structlogger of the component
        log_module: name of the module/component logging the event
        log_event: string describing the log event
        **kwargs: dictionary of additional logging context
    """
    log_level_llm_name = os.environ.get(ENV_LOG_LEVEL_LLM, DEFAULT_LOG_LEVEL_LLM)
    log_level_llm = logging.getLevelName(log_level_llm_name.upper())

    module_env_variable = ENV_LOG_LEVEL_LLM_MODULE_NAMES.get(
        log_module, "LOG_LEVEL_LLM_" + log_module.upper()
    )
    log_level_llm_module_name = os.environ.get(
        module_env_variable, DEFAULT_LOG_LEVEL_LLM
    )
    log_level_llm_module = logging.getLevelName(log_level_llm_module_name.upper())

    # log at the highest specified level, e.g. max(DEBUG=10, INFO=20)
    log_level = max(log_level_llm, log_level_llm_module)

    logger.log(log_level, log_event, **kwargs)



from dataclasses import dataclass, field, asdict
from typing import Dict, List, Optional


@dataclass
class LLMUsage:
    prompt_tokens: int
    """Number of prompt tokens used to generate completion."""

    completion_tokens: int
    """Number of generated tokens."""

    total_tokens: int = field(init=False)
    """Total number of used tokens."""

    def __post_init__(self) -> None:
        self.total_tokens = self.prompt_tokens + self.completion_tokens

    def to_dict(self) -> dict:
        """Converts the LLMUsage dataclass instance into a dictionary."""
        return asdict(self)


@dataclass
class LLMResponse:
    id: str
    """A unique identifier for the completion."""

    choices: List[str]
    """The list of completion choices the model generated for the input prompt."""

    created: int
    """The Unix timestamp (in seconds) of when the completion was created."""

    model: Optional[str] = None
    """The model used for completion."""

    usage: Optional[LLMUsage] = None
    """An optional details about the token usage for the API call."""

    additional_info: Optional[Dict] = None
    """Optional dictionary for storing additional information related to the
    completion that may not be covered by other fields."""

    def to_dict(self) -> dict:
        """Converts the LLMResponse dataclass instance into a dictionary."""
        result = asdict(self)
        if self.usage:
            result["usage"] = self.usage.to_dict()
        return result



from dataclasses import dataclass
from rasa.dialogue_understanding.commands import Command


@dataclass
class FreeFormAnswerCommand(Command):
    """A command to indicate a free-form answer by the bot."""

    pass



import sys
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Text,
    Type,
    TypeVar,
    TYPE_CHECKING,
    Union,
    cast,
)
import json

import structlog

import rasa.shared.utils.io
from rasa.shared.constants import (
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG,
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY,
    PROVIDER_CONFIG_KEY,
)
from rasa.shared.core.events import BotUttered, UserUttered
from rasa.shared.core.slots import Slot, BooleanSlot, CategoricalSlot
from rasa.shared.engine.caching import (
    get_local_cache_location,
)
from rasa.shared.exceptions import (
    FileIOException,
    FileNotFoundException,
    ProviderClientValidationError,
)
from rasa.shared.providers._configs.azure_openai_client_config import (
    is_azure_openai_config,
)
from rasa.shared.providers._configs.huggingface_local_embedding_client_config import (
    is_huggingface_local_config,
)
from rasa.shared.providers._configs.openai_client_config import is_openai_config
from rasa.shared.providers._configs.self_hosted_llm_client_config import (
    is_self_hosted_config,
)
from rasa.shared.providers.embedding.embedding_client import EmbeddingClient
from rasa.shared.providers.llm.llm_client import LLMClient
from rasa.shared.providers.mappings import (
    get_llm_client_from_provider,
    AZURE_OPENAI_PROVIDER,
    OPENAI_PROVIDER,
    SELF_HOSTED_PROVIDER,
    get_embedding_client_from_provider,
    HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER,
    get_client_config_class_from_provider,
)

if TYPE_CHECKING:
    from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()

USER = "USER"

AI = "AI"

DEFAULT_OPENAI_GENERATE_MODEL_NAME = "gpt-3.5-turbo"

DEFAULT_OPENAI_CHAT_MODEL_NAME = "gpt-3.5-turbo"

DEFAULT_OPENAI_CHAT_MODEL_NAME_ADVANCED = "gpt-4-0613"

DEFAULT_OPENAI_EMBEDDING_MODEL_NAME = "text-embedding-ada-002"

DEFAULT_OPENAI_TEMPERATURE = 0.7

DEFAULT_OPENAI_MAX_GENERATED_TOKENS = 256

DEFAULT_MAX_USER_INPUT_CHARACTERS = 420

# Placeholder messages used in the transcript for
# instances where user input results in an error
ERROR_PLACEHOLDER = {
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG: "[User sent really long message]",
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY: "",
    "default": "[User input triggered an error]",
}

_Factory_F = TypeVar(
    "_Factory_F",
    bound=Callable[[Dict[str, Any], Dict[str, Any]], Union[EmbeddingClient, LLMClient]],
)
_CombineConfigs_F = TypeVar(
    "_CombineConfigs_F",
    bound=Callable[[Dict[str, Any], Dict[str, Any]], Dict[str, Any]],
)


def _compute_hash_for_cache_from_configs(
    config_x: Dict[str, Any], config_y: Dict[str, Any]
) -> int:
    """Get a unique hash of the default and custom configs."""
    return hash(
        json.dumps(config_x, sort_keys=True) + json.dumps(config_y, sort_keys=True)
    )


def _retrieve_from_cache(
    cache: Dict[int, Any], unique_hash: int, function: Callable, function_kwargs: dict
) -> Any:
    """Retrieve the value from the cache if it exists. If it does not exist, cache it"""
    if unique_hash in cache:
        return cache[unique_hash]
    else:
        return_value = function(**function_kwargs)
        cache[unique_hash] = return_value
        return return_value


def _cache_factory(function: _Factory_F) -> _Factory_F:
    """Memoize the factory methods based on the arguments."""
    cache: Dict[int, Union[EmbeddingClient, LLMClient]] = {}

    @wraps(function)
    def factory_method_wrapper(
        config_x: Dict[str, Any], config_y: Dict[str, Any]
    ) -> Union[EmbeddingClient, LLMClient]:
        # Get a unique hash of the default and custom configs.
        unique_hash = _compute_hash_for_cache_from_configs(config_x, config_y)
        return _retrieve_from_cache(
            cache=cache,
            unique_hash=unique_hash,
            function=function,
            function_kwargs={"custom_config": config_x, "default_config": config_y},
        )

    def clear_cache() -> None:
        cache.clear()
        structlogger.debug(
            "Cleared cache for factory method",
            function_name=function.__name__,
        )

    setattr(factory_method_wrapper, "clear_cache", clear_cache)
    return cast(_Factory_F, factory_method_wrapper)


def _cache_combine_custom_and_default_configs(
    function: _CombineConfigs_F,
) -> _CombineConfigs_F:
    """Memoize the combine_custom_and_default_config method based on the arguments."""
    cache: Dict[int, dict] = {}

    @wraps(function)
    def combine_configs_wrapper(
        config_x: Dict[str, Any], config_y: Dict[str, Any]
    ) -> dict:
        # Get a unique hash of the default and custom configs.
        unique_hash = _compute_hash_for_cache_from_configs(config_x, config_y)
        return _retrieve_from_cache(
            cache=cache,
            unique_hash=unique_hash,
            function=function,
            function_kwargs={"custom_config": config_x, "default_config": config_y},
        )

    def clear_cache() -> None:
        cache.clear()
        structlogger.debug(
            "Cleared cache for combine_custom_and_default_config method",
            function_name=function.__name__,
        )

    setattr(combine_configs_wrapper, "clear_cache", clear_cache)
    return cast(_CombineConfigs_F, combine_configs_wrapper)


def tracker_as_readable_transcript(
    tracker: "DialogueStateTracker",
    human_prefix: str = USER,
    ai_prefix: str = AI,
    max_turns: Optional[int] = 20,
) -> str:
    """Creates a readable dialogue from a tracker.

    Args:
        tracker: the tracker to convert
        human_prefix: the prefix to use for human utterances
        ai_prefix: the prefix to use for ai utterances
        max_turns: the maximum number of turns to include in the transcript

    Example:
        >>> tracker = Tracker(
        ...     sender_id="test",
        ...     slots=[],
        ...     events=[
        ...         UserUttered("hello"),
        ...         BotUttered("hi"),
        ...     ],
        ... )
        >>> tracker_as_readable_transcript(tracker)
        USER: hello
        AI: hi

    Returns:
    A string representing the transcript of the tracker
    """
    transcript = []

    # using `applied_events` rather than `events` means that only events after the
    # most recent `Restart` or `SessionStarted` are included in the transcript
    for event in tracker.applied_events():
        if isinstance(event, UserUttered):
            if event.has_triggered_error:
                first_error = event.error_commands[0]
                error_type = first_error.get("error_type")
                message = ERROR_PLACEHOLDER.get(
                    error_type, ERROR_PLACEHOLDER["default"]
                )
            else:
                message = sanitize_message_for_prompt(event.text)
            transcript.append(f"{human_prefix}: {message}")

        elif isinstance(event, BotUttered):
            transcript.append(f"{ai_prefix}: {sanitize_message_for_prompt(event.text)}")

    if max_turns:
        transcript = transcript[-max_turns:]

    return "\n".join(transcript)


def sanitize_message_for_prompt(text: Optional[str]) -> str:
    """Removes new lines from a string.

    Args:
        text: the text to sanitize

    Returns:
    A string with new lines removed.
    """
    return text.replace("\n", " ") if text else ""


@_cache_combine_custom_and_default_configs
def combine_custom_and_default_config(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> Dict[Text, Any]:
    """Merges the given llm config with the default config.

    This method guarantees that the provider is set and all the deprecated keys are
    resolved. Hence, produces only a valid client config.

    Only uses the default configuration arguments, if the type set in the
    custom config matches the type in the default config. Otherwise, only
    the custom config is used.

    Args:
        custom_config: The custom config containing values to overwrite defaults
        default_config: The default config.

    Returns:
        The merged config.
    """
    if custom_config is None:
        return default_config.copy()

    # Get the provider from the custom config.
    custom_config_provider = get_provider_from_config(custom_config)
    # We expect the provider to be set in the default configs of all Rasa components.
    default_config_provider = default_config[PROVIDER_CONFIG_KEY]

    if (
        custom_config_provider is not None
        and custom_config_provider != default_config_provider
    ):
        # Get the provider-specific config class
        client_config_clazz = get_client_config_class_from_provider(
            custom_config_provider
        )
        # Checks for deprecated keys, resolves aliases and returns a valid config.
        # This is done to ensure that the custom config is valid.
        return client_config_clazz.from_dict(custom_config).to_dict()

    # If the provider is the same in both configs
    # OR provider is not specified in the custom config
    # perform MERGE by overriding the default config keys and values
    # with custom config keys and values.
    merged_config = {**default_config.copy(), **custom_config.copy()}
    # Check for deprecated keys, resolve aliases and return a valid config.
    # This is done to ensure that the merged config is valid.
    default_config_clazz = get_client_config_class_from_provider(
        default_config_provider
    )
    return default_config_clazz.from_dict(merged_config).to_dict()


def get_provider_from_config(config: dict) -> Optional[str]:
    """Try to get the provider from the passed llm/embeddings configuration.
    If no provider can be found, return None.
    """
    if not config:
        return None
    if is_self_hosted_config(config):
        return SELF_HOSTED_PROVIDER
    elif is_azure_openai_config(config):
        return AZURE_OPENAI_PROVIDER
    elif is_openai_config(config):
        return OPENAI_PROVIDER
    elif is_huggingface_local_config(config):
        return HUGGINGFACE_LOCAL_EMBEDDING_PROVIDER
    else:
        return config.get(PROVIDER_CONFIG_KEY)


def ensure_cache() -> None:
    """Ensures that the cache is initialized."""
    import litellm

    # Ensure the cache directory exists
    cache_location = get_local_cache_location() / "rasa-llm-cache"
    cache_location.mkdir(parents=True, exist_ok=True)

    # Set diskcache as a caching option
    litellm.cache = litellm.Cache(type="disk", disk_cache_dir=cache_location)


@_cache_factory
def llm_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> LLMClient:
    """Creates an LLM from the given config.

    Args:
        custom_config: The custom config  containing values to overwrite defaults
        default_config: The default config.

    Returns:
        Instantiated LLM based on the configuration.
    """
    config = combine_custom_and_default_config(custom_config, default_config)

    ensure_cache()

    client_clazz: Type[LLMClient] = get_llm_client_from_provider(
        config[PROVIDER_CONFIG_KEY]
    )
    client = client_clazz.from_config(config)
    return client


@_cache_factory
def embedder_factory(
    custom_config: Optional[Dict[str, Any]], default_config: Dict[str, Any]
) -> EmbeddingClient:
    """Creates an Embedder from the given config.

    Args:
        custom_config: The custom config containing values to overwrite defaults
        default_config: The default config.


    Returns:
        Instantiated Embedder based on the configuration.
    """
    config = combine_custom_and_default_config(custom_config, default_config)

    ensure_cache()

    client_clazz: Type[EmbeddingClient] = get_embedding_client_from_provider(
        config[PROVIDER_CONFIG_KEY]
    )
    client = client_clazz.from_config(config)
    return client


def get_prompt_template(
    jinja_file_path: Optional[Text], default_prompt_template: Text
) -> Text:
    """Returns the jinja template.

    Args:
        jinja_file_path: the path to the jinja file
        default_prompt_template: the default prompt template

    Returns:
        The prompt template.
    """
    try:
        if jinja_file_path is not None:
            return rasa.shared.utils.io.read_file(jinja_file_path)
    except (FileIOException, FileNotFoundException):
        structlogger.warning(
            "Failed to read custom prompt template. Using default template instead.",
            jinja_file_path=jinja_file_path,
        )
    return default_prompt_template


def allowed_values_for_slot(slot: Slot) -> Union[str, None]:
    """Get the allowed values for a slot."""
    if isinstance(slot, BooleanSlot):
        return str([True, False])
    if isinstance(slot, CategoricalSlot):
        return str([v for v in slot.values if v != "__other__"])
    else:
        return None


def try_instantiate_llm_client(
    custom_llm_config: Optional[Dict],
    default_llm_config: Optional[Dict],
    log_source_function: str,
    log_source_component: str,
) -> None:
    """Validate llm configuration."""
    try:
        llm_factory(custom_llm_config, default_llm_config)
    except (ProviderClientValidationError, ValueError) as e:
        structlogger.error(
            f"{log_source_function}.llm_instantiation_failed",
            event_info=(
                f"Unable to create the LLM client for component - "
                f"{log_source_component}. Please make sure you specified the required "
                f"environment variables and configuration keys."
            ),
            error=e,
        )
        sys.exit(1)


def try_instantiate_embedder(
    custom_embeddings_config: Optional[Dict],
    default_embeddings_config: Optional[Dict],
    log_source_function: str,
    log_source_component: str,
) -> EmbeddingClient:
    """Validate embeddings configuration."""
    try:
        return embedder_factory(custom_embeddings_config, default_embeddings_config)
    except (ProviderClientValidationError, ValueError) as e:
        structlogger.error(
            f"{log_source_function}.embedder_instantiation_failed",
            event_info=(
                f"Unable to create the Embedding client for component - "
                f"{log_source_component}. Please make sure you specified the required "
                f"environment variables and configuration keys."
            ),
            error=e,
        )
        sys.exit(1)



from typing import Any, Dict, Optional, Text, List

import structlog
from rasa.dialogue_understanding.stack.frames.flow_stack_frame import (
    FlowStackFrameType,
    UserFlowStackFrame,
)
from rasa.dialogue_understanding.stack.utils import top_user_flow_frame
from rasa.core.actions import action
from rasa.core.channels import OutputChannel
from rasa.shared.constants import FLOW_PREFIX

from rasa.shared.core.domain import Domain
from rasa.shared.core.events import (
    ActiveLoop,
    Event,
    FlowInterrupted,
    SlotSet,
)
from rasa.core.nlg import NaturalLanguageGenerator
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger(__name__)


class ActionTriggerFlow(action.Action):
    """Action which triggers a flow by putting it on the dialogue stack."""

    def __init__(self, flow_action_name: Text) -> None:
        """Creates a `ActionTriggerFlow`.

        Args:
            flow_action_name: Name of the flow.
        """
        super().__init__()

        if not flow_action_name.startswith(FLOW_PREFIX):
            raise ValueError(
                f"Flow action name '{flow_action_name}' needs to start with "
                f"'{FLOW_PREFIX}'."
            )

        self._flow_name = flow_action_name[len(FLOW_PREFIX) :]
        self._flow_action_name = flow_action_name

    def name(self) -> Text:
        """Return the flow name."""
        return self._flow_action_name

    def create_events_to_start_flow(self, tracker: DialogueStateTracker) -> List[Event]:
        """Create events to start the flow.

        Args:
            tracker: The tracker to start the flow on.

        Returns:
            The event to start the flow."""
        stack = tracker.stack
        events: List[Event] = []

        frame_type = FlowStackFrameType.REGULAR

        if not stack.is_empty():
            frame_type = FlowStackFrameType.INTERRUPT
            top_user_frame = top_user_flow_frame(stack)
            if top_user_frame is not None:
                events.append(
                    FlowInterrupted(top_user_frame.flow_id, top_user_frame.step_id)
                )

        stack.push(
            UserFlowStackFrame(
                flow_id=self._flow_name,
                frame_type=frame_type,
            )
        )
        return events + tracker.create_stack_updated_events(stack)

    def create_events_to_set_flow_slots(self, metadata: Dict[str, Any]) -> List[Event]:
        """Create events to set the flow slots.

        Set additional slots to prefill information for the flow.

        Args:
            metadata: The metadata to set the slots from.

        Returns:
            The events to set the flow slots.
        """
        slots_to_be_set = metadata.get("slots", {}) if metadata else {}
        return [SlotSet(key, value) for key, value in slots_to_be_set.items()]

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Trigger the flow."""
        events: List[Event] = self.create_events_to_start_flow(tracker)
        events.extend(self.create_events_to_set_flow_slots(metadata))

        if tracker.active_loop_name:
            # end any active loop to ensure we are progressing the started flow
            events.append(ActiveLoop(None))

        return events



import asyncio
import logging
import os
import platform
import uuid
import warnings
from asyncio import AbstractEventLoop
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Text,
    Tuple,
    Union,
)

from sanic import Sanic
from sanic.worker.loader import AppLoader

import rasa.core.utils
import rasa.shared.utils.common
import rasa.shared.utils.io
import rasa.utils
import rasa.utils.common
import rasa.utils.io
from rasa import server, telemetry
from rasa.constants import ENV_SANIC_BACKLOG
from rasa.core import agent, channels, constants
from rasa.core.agent import Agent
from rasa.core.channels import console
from rasa.core.channels.channel import InputChannel
from rasa.core.utils import AvailableEndpoints
from rasa.nlu.persistor import StorageType
from rasa.plugin import plugin_manager
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.yaml import read_config_file
from rasa.utils import licensing

logger = logging.getLogger()  # get the root logger


def create_http_input_channels(
    channel: Optional[Text], credentials_file: Optional[Text]
) -> List["InputChannel"]:
    """Instantiate the chosen input channel."""
    if credentials_file:
        all_credentials = read_config_file(credentials_file)
    else:
        all_credentials = {}

    if channel:
        if len(all_credentials) > 1:
            logger.info(
                "Connecting to channel '{}' which was specified by the "
                "'--connector' argument. Any other channels will be ignored. "
                "To connect to all given channels, omit the '--connector' "
                "argument.".format(channel)
            )
        return [_create_single_channel(channel, all_credentials.get(channel))]
    else:
        return [_create_single_channel(c, k) for c, k in all_credentials.items()]


def _create_single_channel(channel: Text, credentials: Dict[Text, Any]) -> Any:
    from rasa.core.channels import BUILTIN_CHANNELS

    if channel in BUILTIN_CHANNELS:
        return BUILTIN_CHANNELS[channel].from_credentials(credentials)
    else:
        # try to load channel based on class name
        try:
            input_channel_class = rasa.shared.utils.common.class_from_module_path(
                channel
            )
            return input_channel_class.from_credentials(credentials)
        except (AttributeError, ImportError):
            raise RasaException(
                f"Failed to find input channel class for '{channel}'. Unknown "
                f"input channel. Check your credentials configuration to "
                f"make sure the mentioned channel is not misspelled. "
                f"If you are creating your own channel, make sure it "
                f"is a proper name of a class in a module."
            )


def _create_app_without_api(cors: Optional[Union[Text, List[Text]]] = None) -> Sanic:
    app = Sanic("rasa_core_no_api", configure_logging=False)

    # Reset Sanic warnings filter that allows the triggering of Sanic warnings
    warnings.filterwarnings("ignore", category=DeprecationWarning, module=r"sanic.*")

    server.add_root_route(app)
    server.configure_cors(app, cors)
    return app


def _is_apple_silicon_system() -> bool:
    # check if the system is MacOS
    if platform.system().lower() != "darwin":
        return False
    # check for arm architecture, indicating apple silicon
    return platform.machine().startswith("arm") or os.uname().machine.startswith("arm")


def configure_app(
    input_channels: Optional[List["InputChannel"]] = None,
    cors: Optional[Union[Text, List[Text], None]] = None,
    auth_token: Optional[Text] = None,
    enable_api: bool = True,
    response_timeout: int = constants.DEFAULT_RESPONSE_TIMEOUT,
    jwt_secret: Optional[Text] = None,
    jwt_private_key: Optional[Text] = None,
    jwt_method: Optional[Text] = None,
    route: Optional[Text] = "/webhooks/",
    port: int = constants.DEFAULT_SERVER_PORT,
    endpoints: Optional[AvailableEndpoints] = None,
    log_file: Optional[Text] = None,
    conversation_id: Optional[Text] = uuid.uuid4().hex,
    use_syslog: bool = False,
    syslog_address: Optional[Text] = None,
    syslog_port: Optional[int] = None,
    syslog_protocol: Optional[Text] = None,
    request_timeout: Optional[int] = None,
    server_listeners: Optional[List[Tuple[Callable, Text]]] = None,
    use_uvloop: Optional[bool] = True,
    keep_alive_timeout: int = constants.DEFAULT_KEEP_ALIVE_TIMEOUT,
) -> Sanic:
    """Run the agent."""
    rasa.core.utils.configure_file_logging(
        logger, log_file, use_syslog, syslog_address, syslog_port, syslog_protocol
    )

    if enable_api:
        loader = AppLoader(
            factory=partial(
                server.create_app,
                cors_origins=cors,
                auth_token=auth_token,
                response_timeout=response_timeout,
                jwt_secret=jwt_secret,
                jwt_private_key=jwt_private_key,
                jwt_method=jwt_method,
                endpoints=endpoints,
            )
        )
    else:
        loader = AppLoader(factory=partial(_create_app_without_api, cors))

    app = loader.load()
    app.config.KEEP_ALIVE_TIMEOUT = keep_alive_timeout

    if _is_apple_silicon_system() or not use_uvloop:
        app.config.USE_UVLOOP = False
        # some library still sets the loop to uvloop, even if disabled for sanic
        # using uvloop leads to breakingio errors, see
        # https://rasahq.atlassian.net/browse/ENG-667
        asyncio.set_event_loop_policy(None)

    if input_channels:
        channels.channel.register(input_channels, app, route=route)
    else:
        input_channels = []

    if logger.isEnabledFor(logging.DEBUG):
        rasa.core.utils.list_routes(app)

    @app.main_process_start
    async def configure_async_logging(running_app: Sanic) -> None:
        if logger.isEnabledFor(logging.DEBUG):
            rasa.utils.io.enable_async_loop_debugging(asyncio.get_event_loop())

    if "cmdline" in {c.name() for c in input_channels}:

        @app.after_server_start
        async def run_cmdline_io(running_app: Sanic) -> None:
            """Small wrapper to shut down the server once cmd io is done."""
            await console.record_messages(
                server_url=constants.DEFAULT_SERVER_FORMAT.format("http", port),
                sender_id=conversation_id,
                request_timeout=request_timeout,
            )

            logger.info("Killing Sanic server now.")
            running_app.stop()  # kill the sanic server

    @app.after_server_stop
    async def after_server_stop(running_app: Sanic) -> None:
        plugin_manager().hook.after_server_stop()

    if server_listeners:
        for listener, event in server_listeners:
            app.register_listener(listener, event)

    return app


def serve_application(
    model_path: Optional[Text] = None,
    channel: Optional[Text] = None,
    interface: Optional[Text] = constants.DEFAULT_SERVER_INTERFACE,
    port: int = constants.DEFAULT_SERVER_PORT,
    credentials: Optional[Text] = None,
    cors: Optional[Union[Text, List[Text]]] = None,
    auth_token: Optional[Text] = None,
    enable_api: bool = True,
    response_timeout: int = constants.DEFAULT_RESPONSE_TIMEOUT,
    jwt_secret: Optional[Text] = None,
    jwt_private_key: Optional[Text] = None,
    jwt_method: Optional[Text] = None,
    endpoints: Optional[AvailableEndpoints] = None,
    remote_storage: Optional[StorageType] = None,
    log_file: Optional[Text] = None,
    ssl_certificate: Optional[Text] = None,
    ssl_keyfile: Optional[Text] = None,
    ssl_ca_file: Optional[Text] = None,
    ssl_password: Optional[Text] = None,
    conversation_id: Optional[Text] = uuid.uuid4().hex,
    use_syslog: Optional[bool] = False,
    syslog_address: Optional[Text] = None,
    syslog_port: Optional[int] = None,
    syslog_protocol: Optional[Text] = None,
    request_timeout: Optional[int] = None,
    server_listeners: Optional[List[Tuple[Callable, Text]]] = None,
) -> None:
    """Run the API entrypoint."""
    if not channel and not credentials:
        channel = "cmdline"

    input_channels = create_http_input_channels(channel, credentials)

    app = configure_app(
        input_channels,
        cors,
        auth_token,
        enable_api,
        response_timeout,
        jwt_secret,
        jwt_private_key,
        jwt_method,
        port=port,
        endpoints=endpoints,
        log_file=log_file,
        conversation_id=conversation_id,
        use_syslog=use_syslog,
        syslog_address=syslog_address,
        syslog_port=syslog_port,
        syslog_protocol=syslog_protocol,
        request_timeout=request_timeout,
        server_listeners=server_listeners,
    )

    ssl_context = server.create_ssl_context(
        ssl_certificate, ssl_keyfile, ssl_ca_file, ssl_password
    )
    protocol = "https" if ssl_context else "http"

    logger.info(f"Starting Rasa server on {protocol}://{interface}:{port}")

    app.register_listener(
        partial(load_agent_on_start, model_path, endpoints, remote_storage),
        "before_server_start",
    )

    app.register_listener(
        licensing.validate_limited_server_license, "after_server_start"
    )

    app.register_listener(close_resources, "after_server_stop")

    number_of_workers = rasa.core.utils.number_of_sanic_workers(
        endpoints.lock_store if endpoints else None
    )

    telemetry.track_server_start(
        input_channels, endpoints, model_path, number_of_workers, enable_api
    )

    rasa.utils.common.update_sanic_log_level(
        log_file, use_syslog, syslog_address, syslog_port, syslog_protocol
    )

    app.run(
        host=interface,
        port=port,
        ssl=ssl_context,
        backlog=int(os.environ.get(ENV_SANIC_BACKLOG, "100")),
        workers=number_of_workers,
        legacy=True,
    )


# noinspection PyUnusedLocal
async def load_agent_on_start(
    model_path: Text,
    endpoints: AvailableEndpoints,
    remote_storage: Optional[StorageType],
    app: Sanic,
    loop: AbstractEventLoop,
) -> Agent:
    """Load an agent.

    Used to be scheduled on server start
    (hence the `app` and `loop` arguments).
    """
    app.ctx.agent = await agent.load_agent(
        model_path=model_path,
        remote_storage=remote_storage,
        endpoints=endpoints,
        loop=loop,
    )
    logger.info("Rasa server is up and running.")
    return app.ctx.agent


async def close_resources(app: Sanic, _: AbstractEventLoop) -> None:
    """Gracefully closes resources when shutting down server.

    Args:
        app: The Sanic application.
        _: The current Sanic worker event loop.
    """
    current_agent = getattr(app.ctx, "agent", None)
    if not current_agent:
        logger.debug("No agent found when shutting down server.")
        return

    event_broker = current_agent.tracker_store.event_broker
    if event_broker:
        await event_broker.close()
